"""This file implements the gym environment of minitaur standing with four legs.

"""
import math
import operator
import random
import time
from gym import spaces
import numpy as np
from pybullet_envs.minitaur.envs import minitaur_gym_env
from pybullet_envs.minitaur.envs import minitaur_logging

INIT_EXTENSION_POS = 2.0
INIT_SWING_POS = 0.0
DESIRED_PITCH = 0
NUM_LEGS = 4
NUM_MOTORS = 2 * NUM_LEGS
STEP_PERIOD = 1.0 / 3.0  # Three steps per second.
STEP_AMPLITUDE = 0.75
PERTURBATION_TOTAL_STEP = 100
MOVING_FLOOR_TOTAL_STEP = 100
DERPY_V0_URDF_VERSION = minitaur_gym_env.DERPY_V0_URDF_VERSION
RAINBOW_DASH_V0_URDF_VERSION = minitaur_gym_env.RAINBOW_DASH_V0_URDF_VERSION


class MinitaurFourLegStandEnv(minitaur_gym_env.MinitaurGymEnv):
  """The gym environment for the minitaur.

  It simulates the a minitaur standing with four legs. The state space
  include the orientation of the torso, and the action space is the desired
  motor angle for each motor. The reward function is based on how close the
  action to zero and the height of the robot base. It prefers a similar pose to
  the signal while keeping balance.
  """
  metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 166}

  def __init__(self,
               urdf_version=None,
               hard_reset=True,
               remove_default_joint_damping=True,
               control_latency=0.0,
               pd_latency=0.0,
               on_rack=False,
               motor_kp=1.0,
               motor_kd=0.02,
               render=False,
               env_randomizer=None,
               use_angular_velocity_in_observation=False,
               use_motor_angle_in_observation=False,
               control_time_step=0.006,
               action_repeat=6,
               log_path=None):
    """Initialize the minitaur alternating legs gym environment.

    Args:
      urdf_version: [DEFAULT_URDF_VERSION, DERPY_V0_URDF_VERSION,
        RAINBOW_DASH_V0_URDF_VERSION] are allowable
        versions. If None, DEFAULT_URDF_VERSION is used. DERPY_V0_URDF_VERSION
        is the result of first pass system identification for derpy.
        We will have a different URDF and related Minitaur class each time we
        perform system identification. While the majority of the code of the
        class remains the same, some code changes (e.g. the constraint location
        might change). __init__() will choose the right Minitaur class from
        different minitaur modules based on urdf_version.
      hard_reset: Whether to wipe the simulation and load everything when reset
        is called. If set to false, reset just place the minitaur back to start
        position and set its pose to initial configuration.
      remove_default_joint_damping: Whether to remove the default joint damping.
      control_latency: It is the delay in the controller between when an
        observation is made at some point, and when that reading is reported
        back to the Neural Network.
      pd_latency: latency of the PD controller loop. PD calculates PWM based on
        the motor angle and velocity. The latency measures the time between when
        the motor angle and velocity are observed on the microcontroller and
        when the true state happens on the motor. It is typically (0.001-
        0.002s).
      on_rack: Whether to place the minitaur on rack. This is only used to debug
        the walking gait. In this mode, the minitaur's base is hung midair so
        that its walking gait is clearer to visualize.
      motor_kp: The P gain of the motor.
      motor_kd: The D gain of the motor.
      render: Whether to render the simulation.
      env_randomizer: An instance (or a list) of EnvRanzomier(s) that can
        randomize the environment during when env.reset() is called and add
        perturbations when env.step() is called.
      use_angular_velocity_in_observation: Whether to include roll_dot and
        pitch_dot of the base in the observation.
      use_motor_angle_in_observation: Whether to include motor angles in the
        observation.
      control_time_step: The time step between two successive control signals.
      action_repeat: The number of simulation steps before actions are applied.
      log_path: The path to write out logs. For the details of logging, refer to
        minitaur_logging.proto.
    """
    # _swing_offset and _extension_offset is to mimic the motor zero-calibration
    # errors.
    self._swing_offset = np.zeros(NUM_LEGS)
    self._extension_offset = np.zeros(NUM_LEGS)
    self._use_angular_velocity_in_observation = use_motor_angle_in_observation
    self._use_motor_angle_in_observation = use_motor_angle_in_observation
    super(MinitaurFourLegStandEnv,
          self).__init__(urdf_version=urdf_version,
                         control_time_step=control_time_step,
                         action_repeat=action_repeat,
                         remove_default_joint_damping=remove_default_joint_damping,
                         accurate_motor_model_enabled=True,
                         motor_overheat_protection=True,
                         hard_reset=hard_reset,
                         motor_kp=motor_kp,
                         motor_kd=motor_kd,
                         control_latency=control_latency,
                         pd_latency=pd_latency,
                         on_rack=on_rack,
                         render=render,
                         env_randomizer=env_randomizer,
                         reflection=False,
                         log_path=log_path)

    action_dim = 4
    action_low = np.array([-1.0] * action_dim)
    action_high = -action_low
    self.action_space = spaces.Box(action_low, action_high)

    self._cam_dist = 1.0
    self._cam_yaw = 30
    self._cam_pitch = -30
    self._perturbation_magnitude = 0.0
    self._sign = 1.0
    self._cur_ori = [0, 0, 0, 1]
    self._goal_ori = [0, 0, 0, 1]

  def reset(self):
    self.desired_pitch = DESIRED_PITCH
    # In this environment, the actions are
    # [swing leg 1, swing leg 2, swing leg 3, swing leg 4,
    #  extension leg 1, extension leg 2, extension leg 3, extension leg 4]
    init_pose = [
        INIT_SWING_POS + self._swing_offset[0], INIT_SWING_POS + self._swing_offset[1],
        INIT_SWING_POS + self._swing_offset[2], INIT_SWING_POS + self._swing_offset[3],
        INIT_EXTENSION_POS + self._extension_offset[0],
        INIT_EXTENSION_POS + self._extension_offset[1],
        INIT_EXTENSION_POS + self._extension_offset[2],
        INIT_EXTENSION_POS + self._extension_offset[3]
    ]
    initial_motor_angles = self._convert_from_leg_model(init_pose)
    self._pybullet_client.resetBasePositionAndOrientation(0, [0, 0, 0], [0, 0, 0, 1])
    super(MinitaurFourLegStandEnv, self).reset(initial_motor_angles=initial_motor_angles,
                                               reset_duration=0.5)
    return self._get_observation()

  def step(self, action):
    """Step forward the simulation, given the action.

    Args:
      action: A list of desired motor angles for eight motors.

    Returns:
      observations: Roll, pitch of the base, and roll, pitch rate.
      reward: The reward for the current state-action pair.
      done: Whether the episode has ended.
      info: A dictionary that stores diagnostic information.

    Raises:
      ValueError: The action dimension is not the same as the number of motors.
      ValueError: The magnitude of actions is out of bounds.
    """
    if self._is_render:
      # Sleep, otherwise the computation takes less time than real time,
      # which will make the visualization like a fast-forward video.
      time_spent = time.time() - self._last_frame_time
      self._last_frame_time = time.time()
      time_to_sleep = self.control_time_step - time_spent
      if time_to_sleep > 0:
        time.sleep(time_to_sleep)
      base_pos = self.minitaur.GetBasePosition()
      # Keep the previous orientation of the camera set by the user.
      [yaw, pitch, dist] = self._pybullet_client.getDebugVisualizerCamera()[8:11]
      self._pybullet_client.resetDebugVisualizerCamera(dist, yaw, pitch, base_pos)
    action = self._transform_action_to_motor_command(action)
    t = self._env_step_counter % MOVING_FLOOR_TOTAL_STEP
    if t == 0:
      self.seed()
      orientation_x = random.uniform(-0.2, 0.2)
      self.seed()
      orientation_y = random.uniform(-0.2, 0.2)
      _, self._cur_ori = self._pybullet_client.getBasePositionAndOrientation(0)
      self._goal_ori = self._pybullet_client.getQuaternionFromEuler(
          [orientation_x, orientation_y, 0])
    t = float(float(t) / float(MOVING_FLOOR_TOTAL_STEP))
    ori = map(operator.add, [x * (1.0 - t) for x in self._cur_ori],
              [x * t for x in self._goal_ori])
    ori = list(ori)
    print("ori=", ori)
    self._pybullet_client.resetBasePositionAndOrientation(0, [0, 0, 0], ori)
    if self._env_step_counter % PERTURBATION_TOTAL_STEP == 0:
      self._perturbation_magnitude = random.uniform(0.0, 0.0)
      if self._sign < 0.5:
        self._sign = 1.0
      else:
        self._sign = 0.0
    self._pybullet_client.applyExternalForce(
        objectUniqueId=1,
        linkIndex=-1,
        forceObj=[self._sign * self._perturbation_magnitude, 0.0, 0.0],
        posObj=[0.0, 0.0, 0.0],
        flags=self._pybullet_client.LINK_FRAME)
    self.minitaur.Step(action)
    self._env_step_counter += 1
    done = self._termination()
    obs = self._get_true_observation()
    reward = self._reward()
    if self._log_path is not None:
      minitaur_logging.update_episode_proto(self._episode_proto, self.minitaur, action,
                                            self._env_step_counter)
    if done:
      self.minitaur.Terminate()
    return np.array(self._get_observation()), reward, done, {}

  def _convert_from_leg_model(self, leg_pose):
    motor_pose = np.zeros(NUM_MOTORS)
    for i in range(NUM_LEGS):
      motor_pose[2 * i] = leg_pose[NUM_LEGS + i] - (-1)**(i / 2) * leg_pose[i]
      motor_pose[2 * i + 1] = leg_pose[NUM_LEGS + i] + (-1)**(i / 2) * leg_pose[i]
    return motor_pose

  def _signal(self, t):
    initial_pose = np.array([
        INIT_SWING_POS, INIT_SWING_POS, INIT_SWING_POS, INIT_SWING_POS, INIT_EXTENSION_POS,
        INIT_EXTENSION_POS, INIT_EXTENSION_POS, INIT_EXTENSION_POS
    ])
    signal = initial_pose
    return signal

  def _transform_action_to_motor_command(self, action):
    # Add swing_offset and extension_offset to mimick the motor zero-calibration
    # errors.
    real_action = np.array([0.0] * 8)
    real_action[4:8] = action + self._extension_offset
    real_action[0:4] = self._swing_offset
    real_action += self._signal(self.minitaur.GetTimeSinceReset())
    real_action = self._convert_from_leg_model(real_action)
    return real_action

  def is_fallen(self):
    """Decide whether the minitaur has fallen.

    # TODO(yunfeibai): choose the fallen option for force perturbation and
    moving floor, and update the comments.

    If the up directions between the base and the world is large (the dot
    product is smaller than 0.85), or the robot base is lower than 0.24, the
    minitaur is considered fallen.

    Returns:
      Boolean value that indicates whether the minitaur has fallen.
    """
    orientation = self.minitaur.GetBaseOrientation()
    rot_mat = self._pybullet_client.getMatrixFromQuaternion(orientation)
    local_up = rot_mat[6:]
    _, _, height = self.minitaur.GetBasePosition()
    local_global_up_dot_product = np.dot(np.asarray([0, 0, 1]), np.asarray(local_up))
    return local_global_up_dot_product < 0.85 or height < 0.15

  def _reward(self):
    roll, pitch, _ = self.minitaur.GetBaseRollPitchYaw()
    return 1.0 / (0.001 + math.fabs(roll) + math.fabs(pitch))

  def _get_observation(self):
    """Get the true observations of this environment.

    It includes the roll, pitch, roll dot, pitch dot of the base, and the motor
    angles.

    Returns:
      The observation list.
    """
    roll, pitch, _ = self.minitaur.GetBaseRollPitchYaw()
    observation = [roll, pitch]
    if self._use_angular_velocity_in_observation:
      roll_rate, pitch_rate, _ = self.minitaur.GetBaseRollPitchYawRate()
      observation.extend([roll_rate, pitch_rate])
    if self._use_motor_angle_in_observation:
      observation.extend(self.minitaur.GetMotorAngles().tolist())
    self._observation = np.array(observation)
    return self._observation

  def _get_true_observation(self):
    """Get the true observations of this environment.

    It includes the roll, pitch, roll dot, pitch dot of the base, and the motor
    angles.

    Returns:
      The observation list.
    """
    roll, pitch, _ = self.minitaur.GetBaseRollPitchYaw()
    observation = [roll, pitch]
    if self._use_angular_velocity_in_observation:
      roll_rate, pitch_rate, _ = self.minitaur.GetBaseRollPitchYawRate()
      observation.extend([roll_rate, pitch_rate])
    if self._use_motor_angle_in_observation:
      observation.extend(self.minitaur.GetMotorAngles().tolist())

    self._observation = np.array(observation)
    return self._observation

  def _get_observation_upper_bound(self):
    """Get the upper bound of the observation.

    Returns:
      The upper bound of an observation. See GetObservation() for the details
        of each element of an observation.
    """
    upper_bound = [2 * math.pi] * 2  # Roll, pitch the base.
    if self._use_angular_velocity_in_observation:
      upper_bound.extend([2 * math.pi / self._time_step] * 2)
    if self._use_motor_angle_in_observation:
      upper_bound.extend([2 * math.pi] * 8)
    return np.array(upper_bound)

  def _get_observation_lower_bound(self):
    lower_bound = -self._get_observation_upper_bound()
    return lower_bound

  def set_swing_offset(self, value):
    """Set the swing offset of each leg.

    It is to mimic the motor zero-calibration errors.

    Args:
      value: A list of four values.
    """
    self._swing_offset = value

  def set_extension_offset(self, value):
    """Set the extension offset of each leg.

    It is to mimic the motor zero-calibration errors.

    Args:
      value: A list of four values.
    """
    self._extension_offset = value

  def set_desired_pitch(self, value):
    """Set the desired pitch of the base, which is a user input.

    Args:
      value: A scalar.
    """
    self.desired_pitch = value



r"""An example to run of the minitaur gym environment with sine gaits.
"""

import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import math
import numpy as np
from pybullet_envs.prediction import pybullet_sim_gym_env
import argparse
import time

from pybullet_envs.prediction import boxstack_pybullet_sim
from gym.wrappers.monitoring import video_recorder


def ResetPoseExample(steps):
  """An example that the minitaur stands still using the reset pose."""

  environment = pybullet_sim_gym_env.PyBulletSimGymEnv(pybullet_sim_factory=boxstack_pybullet_sim,
                                                       debug_visualization=False,
                                                       render=True,
                                                       action_repeat=30)
  action = [math.pi / 2] * 8

  vid = video_recorder.VideoRecorder(env=environment, path="vid.mp4")

  for _ in range(steps):
    print(_)
    startsim = time.time()
    _, _, done, _ = environment.step(action)
    stopsim = time.time()
    startrender = time.time()
    #environment.render(mode='rgb_array')
    vid.capture_frame()
    stoprender = time.time()
    print("env.step ", (stopsim - startsim))
    print("env.render ", (stoprender - startrender))
    if done:
      environment.reset()


def main():
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--env', help='environment ID (0==reset)', type=int, default=0)
  args = parser.parse_args()
  print("--env=" + str(args.env))

  if (args.env == 0):
    ResetPoseExample(steps=1000)


if __name__ == '__main__':
  main()



# Lint as: python3
"""Utilities for robot URDF models."""

from typing import Text

# In the return from pybullet.getJointInfo, the name of the link whose parent is
# that joint.
LINK_NAME_INDEX = 12
# Indication that link_name_to_id should return -1. This is a constant because
# different URDFs use different names for their base links.
BASE_LINK = ""


def link_name_to_id(link_name: Text, robot_id: int, pybullet_client) -> int:
  """Returns the pybullet integer link id corresponding to link_name.

  Args:
    link_name: The name of the link from the URDF. If this is BASE_LINK, returns
      -1, the link id of the base according to pybullet convention.
    robot_id: Integer id of the robot to which the link belongs, as returned by
      pybullet.loadURDF().
    pybullet_client: Client in which the robot is loaded.

  Returns:
    Integer id of the link.

  Raises:
    ValueError if the link_name is not found in the robot.
  """
  if link_name == BASE_LINK:
    return -1
  link_name_list = []
  for i in range(pybullet_client.getNumJoints(robot_id)):
    joint_info = pybullet_client.getJointInfo(robot_id, i)
    link_name_i = joint_info[LINK_NAME_INDEX].decode("UTF-8")
    if link_name_i == link_name:
      return i
    link_name_list.append(link_name_i)
  raise ValueError("Link name '{}' not found in URDF. Options are: {}".format(
      link_name, link_name_list))


def set_collision_filter_group_mask(urdf_id: int, group: int, mask: int,
                                    pybullet_client):
  """Sets the collision filter group and mask to the robot.

  TODO(tingnan): Check if this has side effects with self collision flags
  when loading URDF.

  Args:
    urdf_id: The URDF id as returned by the loadURDF.
    group: The collision group the robot is in. By default, all dynamics objects
      in PyBullet use collision group 1.
    mask: The collision bit mask to use. See go/pybullet for details.
    pybullet_client: The bullet client to use.
  """
  # We includes "-1" for the base link of the URDF.
  for link_id in range(-1, pybullet_client.getNumJoints(urdf_id)):
    pybullet_client.setCollisionFilterGroupMask(urdf_id, link_id, group, mask)



from robot_bases import XmlBasedRobot, MJCFBasedRobot, URDFBasedRobot
import numpy as np
import pybullet
import os
import pybullet_data
from robot_bases import BodyPart


class WalkerBase(MJCFBasedRobot):

  def __init__(self, fn, robot_name, action_dim, obs_dim, power):
    MJCFBasedRobot.__init__(self, fn, robot_name, action_dim, obs_dim)
    self.power = power
    self.camera_x = 0
    self.start_pos_x, self.start_pos_y, self.start_pos_z = 0, 0, 0
    self.walk_target_x = 1e3  # kilometer away
    self.walk_target_y = 0
    self.body_xyz = [0, 0, 0]

  def robot_specific_reset(self, bullet_client):
    self._p = bullet_client
    for j in self.ordered_joints:
      j.reset_current_position(self.np_random.uniform(low=-0.1, high=0.1), 0)

    self.feet = [self.parts[f] for f in self.foot_list]
    self.feet_contact = np.array([0.0 for f in self.foot_list], dtype=np.float32)
    self.scene.actor_introduce(self)
    self.initial_z = None

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    for n, j in enumerate(self.ordered_joints):
      j.set_motor_torque(self.power * j.power_coef * float(np.clip(a[n], -1, +1)))

  def calc_state(self):
    j = np.array([j.current_relative_position() for j in self.ordered_joints],
                 dtype=np.float32).flatten()
    # even elements [0::2] position, scaled to -1..+1 between limits
    # odd elements  [1::2] angular speed, scaled to show -1..+1
    self.joint_speeds = j[1::2]
    self.joints_at_limit = np.count_nonzero(np.abs(j[0::2]) > 0.99)

    body_pose = self.robot_body.pose()
    parts_xyz = np.array([p.pose().xyz() for p in self.parts.values()]).flatten()
    self.body_xyz = (parts_xyz[0::3].mean(), parts_xyz[1::3].mean(), body_pose.xyz()[2]
                    )  # torso z is more informative than mean z
    self.body_real_xyz = body_pose.xyz()
    self.body_rpy = body_pose.rpy()
    z = self.body_xyz[2]
    if self.initial_z == None:
      self.initial_z = z
    r, p, yaw = self.body_rpy
    self.walk_target_theta = np.arctan2(self.walk_target_y - self.body_xyz[1],
                                        self.walk_target_x - self.body_xyz[0])
    self.walk_target_dist = np.linalg.norm(
        [self.walk_target_y - self.body_xyz[1], self.walk_target_x - self.body_xyz[0]])
    angle_to_target = self.walk_target_theta - yaw

    rot_speed = np.array([[np.cos(-yaw), -np.sin(-yaw), 0], [np.sin(-yaw),
                                                             np.cos(-yaw), 0], [0, 0, 1]])
    vx, vy, vz = np.dot(rot_speed,
                        self.robot_body.speed())  # rotate speed back to body point of view

    more = np.array(
        [
            z - self.initial_z,
            np.sin(angle_to_target),
            np.cos(angle_to_target),
            0.3 * vx,
            0.3 * vy,
            0.3 * vz,  # 0.3 is just scaling typical speed into -1..+1, no physical sense here
            r,
            p
        ],
        dtype=np.float32)
    return np.clip(np.concatenate([more] + [j] + [self.feet_contact]), -5, +5)

  def calc_potential(self):
    # progress in potential field is speed*dt, typical speed is about 2-3 meter per second, this potential will change 2-3 per frame (not per second),
    # all rewards have rew/frame units and close to 1.0
    debugmode = 0
    if (debugmode):
      print("calc_potential: self.walk_target_dist")
      print(self.walk_target_dist)
      print("self.scene.dt")
      print(self.scene.dt)
      print("self.scene.frame_skip")
      print(self.scene.frame_skip)
      print("self.scene.timestep")
      print(self.scene.timestep)
    return -self.walk_target_dist / self.scene.dt


class Hopper(WalkerBase):
  foot_list = ["foot"]

  def __init__(self):
    WalkerBase.__init__(self, "hopper.xml", "torso", action_dim=3, obs_dim=15, power=0.75)

  def alive_bonus(self, z, pitch):
    return +1 if z > 0.8 and abs(pitch) < 1.0 else -1


class Walker2D(WalkerBase):
  foot_list = ["foot", "foot_left"]

  def __init__(self):
    WalkerBase.__init__(self, "walker2d.xml", "torso", action_dim=6, obs_dim=22, power=0.40)

  def alive_bonus(self, z, pitch):
    return +1 if z > 0.8 and abs(pitch) < 1.0 else -1

  def robot_specific_reset(self, bullet_client):
    WalkerBase.robot_specific_reset(self, bullet_client)
    for n in ["foot_joint", "foot_left_joint"]:
      self.jdict[n].power_coef = 30.0


class HalfCheetah(WalkerBase):
  foot_list = ["ffoot", "fshin", "fthigh", "bfoot", "bshin",
               "bthigh"]  # track these contacts with ground

  def __init__(self):
    WalkerBase.__init__(self, "half_cheetah.xml", "torso", action_dim=6, obs_dim=26, power=0.90)

  def alive_bonus(self, z, pitch):
    # Use contact other than feet to terminate episode: due to a lot of strange walks using knees
    return +1 if np.abs(pitch) < 1.0 and not self.feet_contact[1] and not self.feet_contact[
        2] and not self.feet_contact[4] and not self.feet_contact[5] else -1

  def robot_specific_reset(self, bullet_client):
    WalkerBase.robot_specific_reset(self, bullet_client)
    self.jdict["bthigh"].power_coef = 120.0
    self.jdict["bshin"].power_coef = 90.0
    self.jdict["bfoot"].power_coef = 60.0
    self.jdict["fthigh"].power_coef = 140.0
    self.jdict["fshin"].power_coef = 60.0
    self.jdict["ffoot"].power_coef = 30.0


class Ant(WalkerBase):
  foot_list = ['front_left_foot', 'front_right_foot', 'left_back_foot', 'right_back_foot']

  def __init__(self):
    WalkerBase.__init__(self, "ant.xml", "torso", action_dim=8, obs_dim=28, power=2.5)

  def alive_bonus(self, z, pitch):
    return +1 if z > 0.26 else -1  # 0.25 is central sphere rad, die if it scrapes the ground


class Humanoid(WalkerBase):
  self_collision = True
  foot_list = ["right_foot", "left_foot"]  # "left_hand", "right_hand"

  def __init__(self):
    WalkerBase.__init__(self,
                        'humanoid_symmetric.xml',
                        'torso',
                        action_dim=17,
                        obs_dim=44,
                        power=0.41)
    # 17 joints, 4 of them important for walking (hip, knee), others may as well be turned off, 17/4 = 4.25

  def robot_specific_reset(self, bullet_client):
    WalkerBase.robot_specific_reset(self, bullet_client)
    self.motor_names = ["abdomen_z", "abdomen_y", "abdomen_x"]
    self.motor_power = [100, 100, 100]
    self.motor_names += ["right_hip_x", "right_hip_z", "right_hip_y", "right_knee"]
    self.motor_power += [100, 100, 300, 200]
    self.motor_names += ["left_hip_x", "left_hip_z", "left_hip_y", "left_knee"]
    self.motor_power += [100, 100, 300, 200]
    self.motor_names += ["right_shoulder1", "right_shoulder2", "right_elbow"]
    self.motor_power += [75, 75, 75]
    self.motor_names += ["left_shoulder1", "left_shoulder2", "left_elbow"]
    self.motor_power += [75, 75, 75]
    self.motors = [self.jdict[n] for n in self.motor_names]
    if self.random_yaw:
      position = [0, 0, 0]
      orientation = [0, 0, 0]
      yaw = self.np_random.uniform(low=-3.14, high=3.14)
      if self.random_lean and self.np_random.randint(2) == 0:
        cpose.set_xyz(0, 0, 1.4)
        if self.np_random.randint(2) == 0:
          pitch = np.pi / 2
          position = [0, 0, 0.45]
        else:
          pitch = np.pi * 3 / 2
          position = [0, 0, 0.25]
        roll = 0
        orientation = [roll, pitch, yaw]
      else:
        position = [0, 0, 1.4]
        orientation = [0, 0, yaw]  # just face random direction, but stay straight otherwise
      self.robot_body.reset_position(position)
      self.robot_body.reset_orientation(orientation)
    self.initial_z = 0.8

  random_yaw = False
  random_lean = False

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    force_gain = 1
    for i, m, power in zip(range(17), self.motors, self.motor_power):
      m.set_motor_torque(float(force_gain * power * self.power * np.clip(a[i], -1, +1)))

  def alive_bonus(self, z, pitch):
    return +2 if z > 0.78 else -1  # 2 here because 17 joints produce a lot of electricity cost just from policy noise, living must be better than dying


def get_cube(_p, x, y, z):
  body = _p.loadURDF(os.path.join(pybullet_data.getDataPath(), "cube_small.urdf"), [x, y, z])
  _p.changeDynamics(body, -1, mass=1.2)  #match Roboschool
  part_name, _ = _p.getBodyInfo(body)
  part_name = part_name.decode("utf8")
  bodies = [body]
  return BodyPart(_p, part_name, bodies, 0, -1)


def get_sphere(_p, x, y, z):
  body = _p.loadURDF(os.path.join(pybullet_data.getDataPath(), "sphere2red_nocol.urdf"), [x, y, z])
  part_name, _ = _p.getBodyInfo(body)
  part_name = part_name.decode("utf8")
  bodies = [body]
  return BodyPart(_p, part_name, bodies, 0, -1)


class HumanoidFlagrun(Humanoid):

  def __init__(self):
    Humanoid.__init__(self)
    self.flag = None

  def robot_specific_reset(self, bullet_client):
    Humanoid.robot_specific_reset(self, bullet_client)
    self.flag_reposition()

  def flag_reposition(self):
    self.walk_target_x = self.np_random.uniform(low=-self.scene.stadium_halflen,
                                                high=+self.scene.stadium_halflen)
    self.walk_target_y = self.np_random.uniform(low=-self.scene.stadium_halfwidth,
                                                high=+self.scene.stadium_halfwidth)
    more_compact = 0.5  # set to 1.0 whole football field
    self.walk_target_x *= more_compact
    self.walk_target_y *= more_compact

    if (self.flag):
      #for b in self.flag.bodies:
      #	print("remove body uid",b)
      #	p.removeBody(b)
      self._p.resetBasePositionAndOrientation(self.flag.bodies[0],
                                              [self.walk_target_x, self.walk_target_y, 0.7],
                                              [0, 0, 0, 1])
    else:
      self.flag = get_sphere(self._p, self.walk_target_x, self.walk_target_y, 0.7)
    self.flag_timeout = 600 / self.scene.frame_skip  #match Roboschool

  def calc_state(self):
    self.flag_timeout -= 1
    state = Humanoid.calc_state(self)
    if self.walk_target_dist < 1 or self.flag_timeout <= 0:
      self.flag_reposition()
      state = Humanoid.calc_state(self)  # caclulate state again, against new flag pos
      self.potential = self.calc_potential()  # avoid reward jump
    return state


class HumanoidFlagrunHarder(HumanoidFlagrun):

  def __init__(self):
    HumanoidFlagrun.__init__(self)
    self.flag = None
    self.aggressive_cube = None
    self.frame = 0

  def robot_specific_reset(self, bullet_client):

    HumanoidFlagrun.robot_specific_reset(self, bullet_client)

    self.frame = 0
    if (self.aggressive_cube):
      self._p.resetBasePositionAndOrientation(self.aggressive_cube.bodies[0], [-1.5, 0, 0.05],
                                              [0, 0, 0, 1])
    else:
      self.aggressive_cube = get_cube(self._p, -1.5, 0, 0.05)
    self.on_ground_frame_counter = 0
    self.crawl_start_potential = None
    self.crawl_ignored_potential = 0.0
    self.initial_z = 0.8

  def alive_bonus(self, z, pitch):
    if self.frame % 30 == 0 and self.frame > 100 and self.on_ground_frame_counter == 0:
      target_xyz = np.array(self.body_xyz)
      robot_speed = np.array(self.robot_body.speed())
      angle = self.np_random.uniform(low=-3.14, high=3.14)
      from_dist = 4.0
      attack_speed = self.np_random.uniform(
          low=20.0, high=30.0)  # speed 20..30 (* mass in cube.urdf = impulse)
      time_to_travel = from_dist / attack_speed
      target_xyz += robot_speed * time_to_travel  # predict future position at the moment the cube hits the robot
      position = [
          target_xyz[0] + from_dist * np.cos(angle), target_xyz[1] + from_dist * np.sin(angle),
          target_xyz[2] + 1.0
      ]
      attack_speed_vector = target_xyz - np.array(position)
      attack_speed_vector *= attack_speed / np.linalg.norm(attack_speed_vector)
      attack_speed_vector += self.np_random.uniform(low=-1.0, high=+1.0, size=(3,))
      self.aggressive_cube.reset_position(position)
      self.aggressive_cube.reset_velocity(linearVelocity=attack_speed_vector)
    if z < 0.8:
      self.on_ground_frame_counter += 1
    elif self.on_ground_frame_counter > 0:
      self.on_ground_frame_counter -= 1
    # End episode if the robot can't get up in 170 frames, to save computation and decorrelate observations.
    self.frame += 1
    return self.potential_leak() if self.on_ground_frame_counter < 170 else -1

  def potential_leak(self):
    z = self.body_xyz[2]  # 0.00 .. 0.8 .. 1.05 normal walk, 1.2 when jumping
    z = np.clip(z, 0, 0.8)
    return z / 0.8 + 1.0  # 1.00 .. 2.0

  def calc_potential(self):
    # We see alive bonus here as a leak from potential field. Value V(s) of a given state equals
    # potential, if it is topped up with gamma*potential every frame. Gamma is assumed 0.99.
    #
    # 2.0 alive bonus if z>0.8, potential is 200, leak gamma=0.99, (1-0.99)*200==2.0
    # 1.0 alive bonus on the ground z==0, potential is 100, leak (1-0.99)*100==1.0
    #
    # Why robot whould stand up: to receive 100 points in potential field difference.
    flag_running_progress = Humanoid.calc_potential(self)

    # This disables crawl.
    if self.body_xyz[2] < 0.8:
      if self.crawl_start_potential is None:
        self.crawl_start_potential = flag_running_progress - self.crawl_ignored_potential
        #print("CRAWL START %+0.1f %+0.1f" % (self.crawl_start_potential, flag_running_progress))
      self.crawl_ignored_potential = flag_running_progress - self.crawl_start_potential
      flag_running_progress = self.crawl_start_potential
    else:
      #print("CRAWL STOP %+0.1f %+0.1f" % (self.crawl_ignored_potential, flag_running_progress))
      flag_running_progress -= self.crawl_ignored_potential
      self.crawl_start_potential = None

    return flag_running_progress + self.potential_leak() * 100



"""The metric reporting system used in our env.

In the gym environment, there are many variables or quantities that can help
researchers to debug, evaluate policy performance. Such quantities may
include: Motor torques for quadrupeds, even when they are controlled in
POSITION mode; Distance to walls while a wheeled robot is navigating the
indoor environment. Often, these metrics are private variables (or can only be
computed from private variables). To expose the user interested metrics
outside the environment's observations, we designed this Metric system that
can be invoked in any modules (robot, tasks, sensors) inside the gym env.
"""

import enum
import logging
from typing import Any, Callable, Dict, Sequence, Text
import gin


@gin.constants_from_enum
class MetricScope(enum.Enum):
  """The supported scope of metrics."""
  # The performance metrics.
  PERFORMANCE = 1,

  # The debug metrics for diagnostic purposes.
  DEBUG = 2,

  # The safety metrics.
  SAFETY = 3,

  # The statistics of episodes in metric format.
  STATISTIC = 4,


class MetricCore(object):
  """Aggregates values of interest to compute statistics."""

  def __init__(
      self,
      name: Text,
      scope: MetricScope,
      single_ep_aggregator: Callable[[Sequence[Any]], Any],
      multi_ep_aggregator: Callable[[Sequence[Any]], Dict[Text, Any]],
  ):
    """Initializes the class.

    Args:
      name: The name of the metric, for example "motor_torques",
        "distance_to_wall", etc. The full name of the metric will have scope
        name in the prefix, i.e. "scope/name".
      scope: The scope of this metric. Most metric should be for DEBUG purpose.
        The scope name will be added to the final name of metric in this way:
          "scope/name", which is standarded format for Tensorboard to group
          named variables.
      single_ep_aggregator: The function to process all aggregated metric
        values. The derived MetricReporter (see below) will implements
        reset_episode() which clears the episode data, and will be called during
        env.reset().
      multi_ep_aggregator: The functions to process multi-episode metric values.
        We assume the inputs to the functions is a list of per episode metric
        values, i.e. each element of the list is the output from the
        single_ep_aggregator.
    """
    self._name = scope.name + "/" + name
    self._single_ep_aggregator = single_ep_aggregator
    self._multi_ep_aggregator = multi_ep_aggregator
    self._episode_data = []

  def report(self, metric_value: Any):
    """Stores the reported metric in the internal buffer.

    Args:
      metric_value: The metric we are interested to report.
    """
    self._episode_data.append(metric_value)


class MetricReporter(MetricCore):
  """Reports the metric using the provided aggregator functions."""

  def get_episode_metric(self) -> Dict[Text, Sequence[Any]]:
    """Processes and returns episode metric values.

    Returns:
      Aggregated metrics for the current episode.
    """
    if self._episode_data:
      return {self._name: self._single_ep_aggregator(self._episode_data)}
    else:
      return {}

  def get_multi_ep_metric(
      self, episodic_metric: Dict[Text, Sequence[Any]]) -> Dict[Text, Any]:
    """Processes the aggregated metrics over many episodes.

    Will not be affected by reset_episode, since we take multi-episode data as
    inputs.

    Args:
      episodic_metric: The per episode metrics. We expect the inputs to contain
        the same key as self._name, and that the value is a list of metric
        values computed using self.get_episode_metrc().

    Returns:
      The processed multi-episode metrics.
    """
    if self._name not in episodic_metric:
      logging.warning(
          "The inputs does not contain the key for the current metric: %s",
          self._name)
      return {}
    outputs = {}
    for key, val in self._multi_ep_aggregator(
        episodic_metric[self._name]).items():
      outputs[self._name + "_" + key] = val
    return outputs

  def reset_episode(self):
    """Clears the episode data stored.

    Will be invoked during env.reset(). This will effect how get_episode_metric
    gets computed.

    """
    self._episode_data = []



"""This file implements the locomotion gym env."""
# pylint: disable=dangerous-default-value

import atexit
import collections
import time

import gin
from gym import spaces
import numpy as np

from pybullet_utils import bullet_client
from pybullet_envs.minitaur.envs import minitaur_logging
from pybullet_envs.minitaur.envs import minitaur_logging_pb2
from pybullet_envs.minitaur.envs_v2 import locomotion_gym_env
from pybullet_envs.minitaur.envs_v2.scenes import scene_base
from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.envs_v2.sensors import space_utils
import pybullet

_ACTION_EPS = 0.01
_NUM_SIMULATION_ITERATION_STEPS = 300
_LOG_BUFFER_LENGTH = 5000


@gin.configurable
class MultiagentMobilityGymEnv(locomotion_gym_env.LocomotionGymEnv):
  """The gym environment for the locomotion tasks."""
  metadata = {
      'render.modes': ['human', 'rgb_array'],
      'video.frames_per_second': 100
  }

  def __init__(self,
               gym_config,
               robot_classes,
               scene: scene_base.SceneBase = scene_base.SceneBase(),
               sensors=None,
               tasks=[],
               global_task=None,
               single_reward=False,
               env_randomizers=None):
    """Initializes the locomotion gym environment.

    Args:
      gym_config: An instance of LocomotionGymConfig.
      robot_classes: A list of robot classes. We provide a class rather than an
        instance due to hard_reset functionality. Parameters are expected to be
        configured with gin.
      scene: An object for managing the robot's surroundings.
      sensors: A list of environmental sensors for observation. This does not
        include on-robot sensors.
      tasks: A list of callable function/class to calculate the reward and
        termination condition. Takes the gym env as the argument when calling.
      global_task: A callable function/class to calculate the reward and
        termination condition for all robots. Takes the gym env as the argument
        when calling.
      single_reward: Whether the environment returns a single reward for all
        agents or a dictionary.
      env_randomizers: A list of EnvRandomizer(s). An EnvRandomizer may
        randomize the physical property of minitaur, change the terrrain during
        reset(), or add perturbation forces during step().

    Raises:
      ValueError: If the num_action_repeat is less than 1, or if number of
        unique robot names do not match the number of robot classes.

    """
    # TODO(sehoonha) split observation and full-state sensors (b/129858214)

    # Makes sure that close() is always called to flush out the logs to the
    # disk.
    atexit.register(self.close)
    self.seed()
    self._gym_config = gym_config
    self._robot_classes = robot_classes
    # Checking uniqueness of names and number of names
    self._scene = scene
    # TODO(sehoonha) change the data structure to dictionary
    # TODO(b/144521291) make sure sensors have their own robot names
    self._sensors = sensors if sensors is not None else list()
    self._log_path = gym_config.log_path
    self._logging = minitaur_logging.MinitaurLogging(self._log_path)
    self._episode_proto = minitaur_logging_pb2.MinitaurEpisode()
    self._data_dir = gym_config.data_dir

    # A dictionary containing the objects in the world other than the robot.
    self._tasks = tasks
    self._global_task = global_task
    self._single_reward = single_reward

    self._env_randomizers = env_randomizers if env_randomizers else []

    # This is a workaround due to the issue in b/130128505#comment5
    for task in self._tasks:
      if isinstance(task, sensor.Sensor):
        self._sensors.append(task)
    if global_task and isinstance(global_task, sensor.Sensor):
      self._sensors.append(global_task)

    # Simulation related parameters.
    self._num_action_repeat = gym_config.simulation_parameters.num_action_repeat
    self._on_rack = gym_config.simulation_parameters.robot_on_rack
    if self._num_action_repeat < 1:
      raise ValueError('number of action repeats should be at least 1.')
    self._sim_time_step = gym_config.simulation_parameters.sim_time_step_s
    self._env_time_step = self._num_action_repeat * self._sim_time_step
    self._env_step_counter = 0

    # TODO(b/73829334): Fix the value of self._num_bullet_solver_iterations.
    self._num_bullet_solver_iterations = int(_NUM_SIMULATION_ITERATION_STEPS /
                                             self._num_action_repeat)
    self._is_render = gym_config.simulation_parameters.enable_rendering

    # The wall-clock time at which the last frame is rendered.
    self._last_frame_time = 0.0
    if self._is_render:
      self._pybullet_client = bullet_client.BulletClient(
          connection_mode=pybullet.GUI)
    else:
      self._pybullet_client = bullet_client.BulletClient()

    if gym_config.simulation_parameters.egl_rendering:
      self._pybullet_client.loadPlugin('eglRendererPlugin')
    self._pybullet_client.enable_cns()

    # If enabled, save the performance profile to profiling_path
    # Use Google Chrome about://tracing to open the file
    if gym_config.profiling_path is not None:
      self._profiling_slot = self._pybullet_client.startStateLogging(
          self._pybullet_client.STATE_LOGGING_PROFILE_TIMINGS,
          gym_config.profiling_path)
      self._profiling_counter = 10
    else:
      self._profiling_slot = -1
    # Build the action space. The action space must be compatible with the
    # robot configuration.

    # The action list contains the name of all actions.
    # TODO(b/144479707): Allow robots to set the action space automatically.

    action_space = collections.OrderedDict()
    for robot_name, action in gym_config.actions.items():
      action_lower_bound = []
      action_upper_bound = []
      for action_scalar in action:
        action_upper_bound.append(action_scalar.upper_bound)
        action_lower_bound.append(action_scalar.lower_bound)
      action_space[robot_name] = spaces.Box(
          np.asarray(action_lower_bound),
          np.asarray(action_upper_bound),
          dtype=np.float32)
    self.action_space = spaces.Dict(action_space)

    # Set the default render options.
    self._camera_dist = gym_config.simulation_parameters.camera_distance
    self._camera_yaw = gym_config.simulation_parameters.camera_yaw
    self._camera_pitch = gym_config.simulation_parameters.camera_pitch
    self._render_width = gym_config.simulation_parameters.render_width
    self._render_height = gym_config.simulation_parameters.render_height

    self._hard_reset = True
    self.reset()

    self._hard_reset = gym_config.simulation_parameters.enable_hard_reset

    # Construct the observation space from the list of sensors. Note that we
    # will reconstruct the observation_space after the robot is created.
    self.observation_space = (
        space_utils.convert_sensors_to_gym_space_dictionary(self.all_sensors()))

  def close(self):
    if self._log_path is not None:
      self._logging.save_episode(self._episode_proto)

    for robot in self._robots:
      robot.Terminate()

  def all_sensors(self):
    """Returns all robot and environmental sensors."""
    robot_sensors = []
    for robot in self._robots:
      robot_sensors += robot.GetAllSensors()
    return robot_sensors + self._sensors

  @gin.configurable('multiagent_mobility_gym_env.MultiagentMobilityGymEnv.reset'
                   )
  def reset(self,
            initial_motor_angles=None,
            reset_duration=1.0,
            reset_visualization_camera=True):
    """Resets the robot's position in the world or rebuild the sim world.

    The simulation world will be rebuilt if self._hard_reset is True.

    Args:
      initial_motor_angles: A list of Floats. The desired joint angles after
        reset. If None, the robot will use its built-in value.
      reset_duration: Float. The time (in seconds) needed to rotate all motors
        to the desired initial values.
      reset_visualization_camera: Whether to reset debug visualization camera on
        reset.

    Returns:
      A numpy array contains the initial observation after reset.
    """
    if self._is_render:
      self._pybullet_client.configureDebugVisualizer(
          self._pybullet_client.COV_ENABLE_RENDERING, 0)

    # Clear the simulation world and rebuild the robot interface.
    if self._hard_reset:
      self._pybullet_client.resetSimulation()
      self._pybullet_client.setPhysicsEngineParameter(
          numSolverIterations=self._num_bullet_solver_iterations)
      self._pybullet_client.setTimeStep(self._sim_time_step)
      self._pybullet_client.setGravity(0, 0, -10)

      # Rebuild the world.
      self._scene.build_scene(self._pybullet_client)

      # Rebuild the robots
      # TODO(b/144545080): Make this scale to more than two agents
      # Have multiple robot classes as a list.
      self._robots = []
      for robot_class in self._robot_classes:

        self._robots.append(
            robot_class(
                pybullet_client=self._pybullet_client,
                # TODO(rosewang): Remove on rack in multiagent acase
                on_rack=self._on_rack))

    # Reset the pose of the robot.
    for robot in self._robots:
      robot.Reset(
          reload_urdf=False,
          default_motor_angles=initial_motor_angles,
          reset_time=reset_duration)

    self._env_step_counter = 0
    self._pybullet_client.resetDebugVisualizerCamera(self._camera_dist,
                                                     self._camera_yaw,
                                                     self._camera_pitch,
                                                     [0, 0, 0])

    # Flush the logs to disc and reinitialize the logging system.
    if self._log_path is not None:
      self._logging.save_episode(self._episode_proto)
      self._episode_proto = minitaur_logging_pb2.MinitaurEpisode()
      minitaur_logging.preallocate_episode_proto(self._episode_proto,
                                                 _LOG_BUFFER_LENGTH,
                                                 self._robots[0])
    self._pybullet_client.setPhysicsEngineParameter(enableConeFriction=0)
    self._env_step_counter = 0
    if reset_visualization_camera:
      self._pybullet_client.resetDebugVisualizerCamera(self._camera_dist,
                                                       self._camera_yaw,
                                                       self._camera_pitch,
                                                       [0, 0, 0])

    self._last_action = {
        robot_name: np.zeros(space.shape)
        for robot_name, space in self.action_space.spaces.items()
    }

    if self._is_render:
      self._pybullet_client.configureDebugVisualizer(
          self._pybullet_client.COV_ENABLE_RENDERING, 1)

    for s in self.all_sensors():
      # set name
      if any([r.name in s.get_name() for r in self.robots]):
        robot = [r for r in self.robots if r.name in s.get_name()][0]
        s.set_robot(robot)

    for task in self._tasks:
      if hasattr(task, 'reset'):
        task.reset(self)
    if self._global_task and hasattr(self._global_task, 'reset'):
      self._global_task.reset(self)

    # Loop over all env randomizers.
    for env_randomizer in self._env_randomizers:
      env_randomizer.randomize_env(self)

    for s in self.all_sensors():
      s.on_reset(self)

    return self._get_observation()

  def get_robot(self, name):
    for robot in self.robots:
      if robot.name == name:
        return robot

  def _reward(self):
    """Returns a list of rewards.

    Returns:
      A list of rewards corresponding to each robot and their task.
    """
    global_reward = 0
    if self._global_task:
      global_reward = self._global_task(self)
    if self._single_reward:  # Needed for tfagents compatibility.
      if self._tasks:
        return min([task(self) + global_reward for task in self._tasks])
      return 0
    else:
      if self._tasks:
        return [task(self) + global_reward for task in self._tasks]
      return [0 for _ in self.robots]

  def step(self, actions):
    """Step forward the simulation, given the actions.

    Args:
      actions: A dictionary of actions for all robots. The action for each robot
        can be joint angles for legged platforms like Laikago, and base
        velocity/steering for kinematic robots such like Fetch.

    Returns:
      observations: The observation dictionary. The keys are the sensor names
        and the values are the sensor readings.
      reward: The reward for the current state-action pair.
      done: Whether the episode has ended.
      info: A dictionary that stores diagnostic information.

    Raises:
      ValueError: The action dimension is not the same as the number of motors.
      ValueError: The magnitude of actions is out of bounds.
    """
    self._last_base_position = [
        robot.GetBasePosition() for robot in self._robots
    ]
    self._last_action = actions

    if self._is_render:
      # Sleep, otherwise the computation takes less time than real time,
      # which will make the visualization like a fast-forward video.
      time_spent = time.time() - self._last_frame_time
      self._last_frame_time = time.time()
      time_to_sleep = self._env_time_step - time_spent
      if time_to_sleep > 0:
        time.sleep(time_to_sleep)
      camera_target = np.mean(
          [robot.GetBasePosition() for robot in self._robots], axis=0)

      # Also keep the previous orientation of the camera set by the user.
      [yaw, pitch,
       dist] = self._pybullet_client.getDebugVisualizerCamera()[8:11]
      self._pybullet_client.resetDebugVisualizerCamera(dist, yaw, pitch,
                                                       camera_target)

    for env_randomizer in self._env_randomizers:
      env_randomizer.randomize_step(self)

    # Stepping broken down into their parts
    for robot in self._robots:
      robot.PreStepPerStep(actions)

    for _ in range(self._num_action_repeat):
      for robot in self._robots:
        robot.PreStepPerActionRepeat(actions)

      self._pybullet_client.stepSimulation()

      for robot in self._robots:
        robot.PostStepPerActionRepeat()

    for robot in self._robots:
      robot.PostStepPerStep()

    if self._profiling_slot >= 0:
      self._profiling_counter -= 1
      if self._profiling_counter == 0:
        self._pybullet_client.stopStateLogging(self._profiling_slot)
    if self._log_path is not None:
      minitaur_logging.update_episode_proto(self._episode_proto,
                                            self._robots[0], actions,
                                            self._env_step_counter)
    reward = self._reward()

    for s in self.all_sensors():
      s.on_step(self)

    for task in self._tasks:
      if hasattr(task, 'update'):
        task.update(self)
    if self._global_task and hasattr(self._global_task, 'update'):
      self._global_task.update(self)

    done = self._termination()
    self._env_step_counter += 1
    if done:
      for robot in self._robots:
        robot.Terminate()
    return self._get_observation(), reward, done, {}

  def render(self, mode='rgb_array'):
    if mode != 'rgb_array':
      raise ValueError('Unsupported render mode:{}'.format(mode))
    base_pos = np.mean([robot.GetBasePosition() for robot in self._robots],
                       axis=0)
    view_matrix = self._pybullet_client.computeViewMatrixFromYawPitchRoll(
        cameraTargetPosition=base_pos,
        distance=self._camera_dist,
        yaw=self._camera_yaw,
        pitch=self._camera_pitch,
        roll=0,
        upAxisIndex=2)
    proj_matrix = self._pybullet_client.computeProjectionMatrixFOV(
        fov=60,
        aspect=float(self._render_width) / self._render_height,
        nearVal=0.1,
        farVal=100.0)
    (_, _, px, _, _) = self._pybullet_client.getCameraImage(
        width=self._render_width,
        height=self._render_height,
        renderer=self._pybullet_client.ER_BULLET_HARDWARE_OPENGL,
        viewMatrix=view_matrix,
        projectionMatrix=proj_matrix)
    rgb_array = np.array(px)
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def _termination(self):
    if not all([robot.is_safe for robot in self._robots]):
      return True

    if self._tasks:
      return (self._global_task and self._global_task.done(self)) or any(
          [task.done(self) for task in self._tasks])

    for s in self.all_sensors():
      s.on_terminate(self)

    return False

  def set_time_step(self, num_action_repeat, sim_step=0.001):
    """Sets the time step of the environment.

    Args:
      num_action_repeat: The number of simulation steps/action repeats to be
        executed when calling env.step().
      sim_step: The simulation time step in PyBullet. By default, the simulation
        step is 0.001s, which is a good trade-off between simulation speed and
        accuracy.

    Raises:
      ValueError: If the num_action_repeat is less than 1.
    """
    if num_action_repeat < 1:
      raise ValueError('number of action repeats should be at least 1.')
    self._sim_time_step = sim_step
    self._num_action_repeat = num_action_repeat
    self._env_time_step = sim_step * num_action_repeat
    self._num_bullet_solver_iterations = (
        _NUM_SIMULATION_ITERATION_STEPS / self._num_action_repeat)
    self._pybullet_client.setPhysicsEngineParameter(
        numSolverIterations=self._num_bullet_solver_iterations)
    self._pybullet_client.setTimeStep(self._sim_time_step)
    for robot in self._robots:
      robot.SetTimeSteps(self._num_action_repeat, self._sim_time_step)

  def get_time_since_reset(self):
    """Get the time passed (in seconds) since the last reset.

    Returns:
      List of time in seconds since the last reset for each robot.
    """
    return self._robots[0].GetTimeSinceReset()

  @property
  def tasks(self):
    return self._tasks

  @property
  def robots(self):
    return self._robots

  @property
  def num_robots(self):
    return len(self._robots)



# Lint as: python3
"""A sensor that measures the acceleration of the robot base."""

from typing import Any, Callable, Sequence, Type, Text, Union

import gin
import gym
import numpy as np

from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.envs_v2.utilities import noise_generators

_ACCELEROMETER_DIM = 3
_DEFAULT_ACCELEROMETER_LOWER_BOUND = (-1, -1, -1)
_DEFAULT_ACCELEROMETER_UPPER_BOUND = (1, 1, 1)


@gin.configurable
class AccelerometerSensor(sensor.Sensor):
  """An Accelerometer sensor."""

  def __init__(
      self,
      name: Text = "Accelerometer",
      dtype: Type[Any] = np.float64,
      lower_bound: Sequence[float] = _DEFAULT_ACCELEROMETER_LOWER_BOUND,
      upper_bound: Sequence[float] = _DEFAULT_ACCELEROMETER_UPPER_BOUND,
      noise_generator: Union[Callable[..., Any],
                             noise_generators.NoiseGenerator] = None,
      sensor_latency: Union[float, Sequence[float]] = 0.0,
  ):
    """Constructs AccelerometerSensor.

    Generates separate IMU value channels as per configuration.

    Args:
      name: the name of the sensor.
      dtype: data type of sensor value.
      lower_bound: The lower bounds of the sensor reading.
      upper_bound: The upper bounds of the sensor reading.
      noise_generator: Used to add noise to the readings.
      sensor_latency: There are two ways to use this expected sensor latency.
        For both methods, the latency should be in the same unit as the sensor
        data timestamp. 1. As a single float number, the observation will be a
        1D array. For real robots, this should be set to 0.0. 2. As a array of
        floats, the observation will be a 2D array based on how long the history
        need to be. Thus, [0.0, 0.1, 0.2] is a history length of 3.

    """
    super().__init__(
        name=name,
        sensor_latency=sensor_latency,
        interpolator_fn=sensor.linear_obs_blender)

    self._noise_generator = noise_generator
    self._dtype = dtype

    if lower_bound is None or upper_bound is None:
      raise ValueError("Must provides bounds for the Accelerometer readings.")

    if len(lower_bound) != _ACCELEROMETER_DIM or len(
        upper_bound) != _ACCELEROMETER_DIM:
      raise ValueError(
          "Bounds must be {} dimensions.".format(_ACCELEROMETER_DIM))

    lower_bound = np.array(lower_bound, dtype=dtype)
    upper_bound = np.array(upper_bound, dtype=dtype)

    self._observation_space = self._stack_space(
        gym.spaces.Box(
            low=np.array(lower_bound, dtype=self._dtype),
            high=np.array(upper_bound, dtype=self._dtype),
            dtype=self._dtype))

  def _get_original_observation(self):
    return self._robot.timestamp, np.array(
        self._robot.base_acceleration_accelerometer, dtype=self._dtype)

  def get_observation(self) -> np.ndarray:
    delayed_observation = super().get_observation()
    if self._noise_generator:
      if callable(self._noise_generator):
        return self._noise_generator(delayed_observation)
      else:
        return self._noise_generator.add_noise(delayed_observation)
    return delayed_observation



from pybullet_utils import bullet_client as bc
from pybullet_utils import urdfEditor as ed
import pybullet
import pybullet_data
import time
import argparse

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--urdf_in', help='urdf file name input', default='test.urdf')
parser.add_argument('--urdf_out', help='urdf file name output', default='out.urdf')
parser.add_argument('--urdf_flags', help='urdf flags', type=int, default=0)



args = parser.parse_args()


p0 = bc.BulletClient(connection_mode=pybullet.DIRECT)

p0.setAdditionalSearchPath(pybullet_data.getDataPath())
body_id = p0.loadURDF(args.urdf_in, flags = args.urdf_flags)
ed0 = ed.UrdfEditor()
ed0.initializeFromBulletBody(body_id, p0._client)
ed0.saveUrdf(args.urdf_out)




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl import app
from absl import flags
from keras.callbacks import ModelCheckpoint
from keras.layers import Dense
from keras.models import Sequential
from keras import optimizers
import numpy

FLAGS = flags.FLAGS

flags.DEFINE_string(
    "input_filename", "data/minitaur_log_latency_0.01.csv", "The name of the input CSV file."
    "Each line in the CSV file will contain the motor position, the "
    "motor speed, action and torques.")


def main(unused_argv):
  # fix random seed for reproducibility
  numpy.random.seed(7)
  # load pima indians dataset
  dataset = numpy.loadtxt(FLAGS.input_filename, delimiter=",")
  # split into input (X) and output (Y) variables
  x = dataset[:, 0:3]
  y = dataset[:, 3]
  print("x=", x)
  print("y=", y)

  # create model
  model = Sequential()
  model.add(Dense(12, input_dim=3, activation="relu"))
  model.add(Dense(8, activation="sigmoid"))
  model.add(Dense(1, activation="linear"))

  # Compile model (use adam or sgd)
  model.compile(loss="mean_squared_error", optimizer="adam", metrics=["mean_squared_error"])

  # checkpoint
  filepath = "/tmp/keras/weights-improvement-{epoch:02d}-{val_loss:.2f}.hdf5"
  checkpoint = ModelCheckpoint(filepath,
                               monitor="val_loss",
                               verbose=1,
                               save_best_only=True,
                               mode="min")
  callbacks_list = [checkpoint]

  # Fit the model
  # model.fit(X, Y, epochs=150, batch_size=10)
  # model.fit(X, Y, epochs=150, batch_size=10, callbacks=callbacks_list)
  model.fit(x,
            y,
            validation_split=0.34,
            epochs=4500,
            batch_size=1024,
            callbacks=callbacks_list,
            verbose=0)

  # evaluate the model
  scores = model.evaluate(x, y)
  print("\n%s: %.2f%%" % (model.metrics_names[1], scores[1] * 100))


if __name__ == "__main__":
  app.run(main)



"""Implements dynamic locomotion gym env that changes based on iteration."""
import gin

import pybullet


# TODO(b/142642890): Make this reset to the initial world state first.
@gin.configurable
def gap_task_curriculum_update(env,
                               num_iter,
                               distance_to_gap_or_hurdle=1.5,
                               initial_gap_length=0.1,
                               max_iterations=500,
                               gap_delta=0.0008):
  """Linearly increase the gap width wrt the iteration number.

  This is specific to BuildSingleGapWorld.

  Args:
    env: An instance of a LocomotionGymEnv.
    num_iter: The training iteration we are on.
    distance_to_gap_or_hurdle: The distance to the gap.
    initial_gap_length: The starting gap length.
    max_iterations: The number of iterations up to which we will modify the
      environment.
    gap_delta: The amount to increase the gap width by for each increase of 1 in
      the iteration.
  """

  gap_length = initial_gap_length + gap_delta * min(max_iterations, num_iter)
  env.task.reset(
      env,
      distance_to_gap_or_hurdle=distance_to_gap_or_hurdle,
      gap_or_hurdle_width=gap_length)


@gin.configurable
def gap_world_curriculum_update(env,
                                num_iter,
                                initial_second_block_x=8.15,
                                max_iterations=500,
                                gap_delta=0.0008):
  """Update the world, linearly increasing gap width wrt iteration number.

  This is specific to SingleGapScene.

  Args:
    env: An instance of a LocomotionGymEnv.
    num_iter: The training iteration we are on.
    initial_second_block_x: The initial x position of the second block.
    max_iterations: The number of iterations up to which we will modify the
      environment.
    gap_delta: The amount to increase the gap width by for each increase of 1 in
      the iteration.
  """

  ground = env.scene.ground_ids
  pos = pybullet.getBasePositionAndOrientation(ground[-1])[0]
  # Linearly increase the gap width to 0.5m by the last iteration.
  next_x = initial_second_block_x + gap_delta * min(max_iterations, num_iter)
  pybullet.resetBasePositionAndOrientation(ground[-1], (next_x, pos[1], pos[2]),
                                           [0, 0, 0, 1])



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to calculate Laikago's pose and motor angles."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import attr

LAIKAGO_DEFAULT_ABDUCTION_ANGLE = 0
LAIKAGO_DEFAULT_HIP_ANGLE = 0.67
LAIKAGO_DEFAULT_KNEE_ANGLE = -1.25


@attr.s
class LaikagoPose(object):
  """Default pose of the Laikago.

    Leg order:
    0 -> Front Right.
    1 -> Front Left.
    2 -> Rear Right.
    3 -> Rear Left.
  """
  abduction_angle_0 = attr.ib(type=float, default=0)
  hip_angle_0 = attr.ib(type=float, default=0)
  knee_angle_0 = attr.ib(type=float, default=0)
  abduction_angle_1 = attr.ib(type=float, default=0)
  hip_angle_1 = attr.ib(type=float, default=0)
  knee_angle_1 = attr.ib(type=float, default=0)
  abduction_angle_2 = attr.ib(type=float, default=0)
  hip_angle_2 = attr.ib(type=float, default=0)
  knee_angle_2 = attr.ib(type=float, default=0)
  abduction_angle_3 = attr.ib(type=float, default=0)
  hip_angle_3 = attr.ib(type=float, default=0)
  knee_angle_3 = attr.ib(type=float, default=0)



# Lint as: python3
"""The configuration parameters for our robots."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import enum
from typing import Sequence, Union
import dataclasses
import gin
import numpy as np


@gin.constants_from_enum
class MotorControlMode(enum.Enum):
  """The supported motor control modes."""
  POSITION = 1,

  # Apply motor torques directly.
  TORQUE = 2,

  # Apply a tuple (q, qdot, kp, kd, tau) for each motor. Here q, qdot are motor
  # position and velocities. kp and kd are PD gains. tau is the additional
  # motor torque. This is the most flexible control mode.
  HYBRID = 3,

  # PWM mode is only availalbe for Minitaur
  PWM = 4


# TODO(b/127675924): Group other parameters in the named attrib class.

# Each hybrid action is a tuple (position, position_gain, velocity,
# velocity_gain, torque)
HYBRID_ACTION_DIMENSION = 5


class HybridActionIndex(enum.Enum):
  # The index of each component within the hybrid action tuple.
  POSITION = 0
  POSITION_GAIN = 1
  VELOCITY = 2
  VELOCITY_GAIN = 3
  TORQUE = 4


@gin.configurable
class MotorLimits(object):
  """The data class for motor limits."""

  def __init__(
      self,
      angle_lower_limits: Union[float, Sequence[float]] = float('-inf'),
      angle_upper_limits: Union[float, Sequence[float]] = float('inf'),
      velocity_lower_limits: Union[float, Sequence[float]] = float('-inf'),
      velocity_upper_limits: Union[float, Sequence[float]] = float('inf'),
      torque_lower_limits: Union[float, Sequence[float]] = float('-inf'),
      torque_upper_limits: Union[float, Sequence[float]] = float('inf'),
  ):
    """Initializes the class."""
    self.angle_lower_limits = angle_lower_limits
    self.angle_upper_limits = angle_upper_limits
    self.velocity_lower_limits = velocity_lower_limits
    self.velocity_upper_limits = velocity_upper_limits
    self.torque_lower_limits = torque_lower_limits
    self.torque_upper_limits = torque_upper_limits


@gin.constants_from_enum
class WheeledRobotControlMode(enum.Enum):
  """The control mode for wheeled robots."""
  # Controls the base of the robot (i.e. in kinematic mode.) or the base wheels
  # using motor commands.
  BASE = 1
  # Controls arm only
  ARM = 2
  # Controls both base and arm
  BASE_AND_ARM = 3
  # Controls both base and head
  BASE_AND_HEAD = 4
  # Controls the non-wheel motors. This include arms and heads.
  BODY = 5
  # Controls all degrees of freedom, i.e. the base and arm/head simultaneously.
  ALL = 6
  # High-level navigation target.
  NAVIGATION_TARGET = 7
  # Individually addressable actions for body joints, with nested dict actions.
  ADDRESSABLE = 8


@dataclasses.dataclass
class TwistActionLimits:
  """The data class for twist action limits.

  Common abbreviations used in variable names suffix:
    mps = Meters per Second
    rps = Radians per Second
  """
  max_linear_mps: float
  min_linear_mps: float
  max_angular_rps: float
  min_angular_rps: float






# Lint as: python3
"""Common functionalities to process point clouds."""

import math
import numpy as np


def distance_map_to_point_cloud(distances, fov, width, height):
  """Converts from a depth map to a point cloud.

  Args:
    distances: An numpy array which has the shape of (height, width) that
      denotes a distance map. The unit is meter.
    fov: The field of view of the camera in the vertical direction. The unit
      is radian.
    width: The width of the image resolution of the camera.
    height: The height of the image resolution of the camera.

  Returns:
    point_cloud: The converted point cloud from the distance map. It is a numpy
      array of shape (height, width, 3).
  """
  f = height / (2 * math.tan(fov / 2.0))
  px = np.tile(np.arange(width), [height, 1])
  x = (2 * (px + 0.5) - width) / f * distances / 2
  py = np.tile(np.arange(height), [width, 1]).T
  y = (2 * (py + 0.5) - height) / f * distances / 2
  point_cloud = np.stack((x, y, distances), axis=-1)
  return point_cloud


def point_cloud_normals(points):
  """Estimates the normals of a point cloud using cross product.

  Args:
    points: a point cloud represented as an 2D array of which has the shape of
      [m, n, 3].

  Returns:
    normals: a 2D array of normal vectors which has the same shape as the input.
  """
  grad = np.gradient(points, axis=(0, 1))  # calculates gradient along x and y
  normal = np.cross(grad[0], grad[1])
  norms = np.linalg.norm(normal, axis=2)[:, :, None]  # norms in shape [m,n,1]
  normalized = normal / (norms + 1e-7)  # adds eps to avoid div by 0
  return normalized


def point_cloud_distances(point, point_cloud):
  """Estimates the distance of a point to every point in a point cloud.

  Args:
    point: a 3D vector representing a point.
    point_cloud: a point cloud represented as an 2D array of which has the shape
      of [m, n, 3].

  Returns:
    distances: a distance field represented as a 2D array [m, n].
  """
  return np.linalg.norm(point_cloud - point, axis=2)



import pybullet as p
import pybullet_data as pd
import time
p.connect(p.GUI)
p.setAdditionalSearchPath(pd.getDataPath())
objects=[]
useMaximalCoordinates=False

if 0:
  collisionShapeId = p.createCollisionShape(shapeType=p.GEOM_SPHERE, radius = 0.1)
  batchPositions = []
  index = 1
  for x in range(2):
    for y in range(1):
      for z in range(1):
        batchPositions.append(arr[index])
        index=index+1
  bodyUids = p.createMultiBody(baseMass=10,
                              baseInertialFramePosition=[0, 0, 0],
                              baseCollisionShapeIndex=collisionShapeId,
                              baseVisualShapeIndex=-1,
                              basePosition=[0, 0, 2],
                              batchPositions=batchPositions,
                              useMaximalCoordinates=useMaximalCoordinates)
  for b in bodyUids:
    objects.append(b)

else:
  for i in range(2):
    for j in range(1):
      for k in range(1):
        ob = p.loadURDF("sphere_1cm.urdf", [0.210050 * i, 0.210050 * j, 1 + 0.210050 * k],globalScaling=20,
                        useMaximalCoordinates=useMaximalCoordinates)
        objects.append(ob)
        p.changeDynamics(ob, -1, activationState=p.ACTIVATION_STATE_ENABLE_SLEEPING, linearDamping=0, angularDamping=0, sleepThreshold=0.05)
        if (i==0):
          p.resetBaseVelocity(ob, [0,0,0], [0,0,0.22])
        
        
timeid = p.addUserDebugText("t=", [0,0,2])
lvelid = p.addUserDebugText("lvel", [0,0,1.8])
avelid = p.addUserDebugText("avel", [0,0,1.6])
t=0
dt=1./240.
while p.isConnected():
  p.stepSimulation()
  t+=dt
  txtid = p.addUserDebugText("t="+str(t), [0,0,2],replaceItemUniqueId=timeid)
  lin, ang = p.getBaseVelocity(ob)
  txtid = p.addUserDebugText("lvel="+"{:.4f}".format(lin[0])+","+"{:.4f}".format(lin[1])+","+"{:.4f}".format(lin[2]), [0,0,1.8],replaceItemUniqueId=lvelid)
  txtid = p.addUserDebugText("avel="+"{:.4f}".format(ang[0])+","+"{:.4f}".format(ang[1])+","+"{:.4f}".format(ang[2]), [0,0,1.6],replaceItemUniqueId=avelid)
  time.sleep(dt)
  


# Lint as: python3
"""The on robot sensor classes."""

import enum
from typing import Any, Callable, Iterable, Sequence, Type, Text, Union

import gin
import gym
import numpy as np

from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.envs_v2.utilities import noise_generators


@gin.constants_from_enum
class IMUChannel(enum.Enum):
  ROLL = 1,
  PITCH = 2,
  YAW = 3,
  ROLL_RATE = 4,
  PITCH_RATE = 5,
  YAW_RATE = 6,


@gin.configurable
class IMUSensor(sensor.Sensor):
  """An IMU sensor."""

  def __init__(
      self,
      name: Text = "IMU",
      dtype: Type[Any] = np.float64,
      channels: Sequence[IMUChannel] = None,
      lower_bound: Union[float, Iterable[float]] = None,
      upper_bound: Union[float, Iterable[float]] = None,
      noise_generator: Union[Callable[..., Any],
                             noise_generators.NoiseGenerator] = None,
      sensor_latency: Union[float, Sequence[float]] = 0.0,
  ):
    """Constructs IMUSensor.

    Generates separate IMU value channels as per configuration.

    Args:
      name: the name of the sensor.
      dtype: data type of sensor value.
      channels: value channels wants to subscribe. Must be members of the
        IMUChannel class.
      lower_bound: The lower bounds of the sensor reading.
      upper_bound: The upper bounds of the sensor reading.
      noise_generator: Used to add noise to the readings.
      sensor_latency: There are two ways to use this expected sensor latency.
        For both methods, the latency should be in the same unit as the sensor
        data timestamp. 1. As a single float number, the observation will be a
        1D array. For real robots, this should be set to 0.0. 2. As a array of
        floats, the observation will be a 2D array based on how long the history
        need to be. Thus, [0.0, 0.1, 0.2] is a history length of 3.

    Raises:
      ValueError: If no IMU channel is provided and no bounds for the channels.
    """
    super().__init__(
        name=name,
        sensor_latency=sensor_latency,
        interpolator_fn=sensor.linear_obs_blender)
    if channels is None:
      raise ValueError("IMU channels are not provided.")
    self._channels = channels
    self._num_channels = len(self._channels)
    self._noise_generator = noise_generator
    self._dtype = dtype

    if lower_bound is None or upper_bound is None:
      raise ValueError("Must provides bounds for the IMU readings.")

    if isinstance(lower_bound, (float, int)):
      lower_bound = np.full(self._num_channels, lower_bound, dtype=dtype)
    else:
      lower_bound = np.array(lower_bound, dtype=dtype)

    if len(lower_bound) != self._num_channels:
      raise ValueError("length of sensor lower bound {lower_bound} does not"
                       " match the number of channels.")

    if isinstance(upper_bound, (float, int)):
      upper_bound = np.full(self._num_channels, upper_bound, dtype=dtype)
    else:
      upper_bound = np.array(upper_bound, dtype=dtype)

    if len(upper_bound) != self._num_channels:
      raise ValueError("length of sensor upper bound {upper_bound} does not"
                       " match the number of channels.")

    self._observation_space = self._stack_space(
        gym.spaces.Box(
            low=np.array(lower_bound, dtype=self._dtype),
            high=np.array(upper_bound, dtype=self._dtype),
            dtype=self._dtype))

  def get_channels(self) -> Sequence[IMUChannel]:
    return self._channels

  def get_num_channels(self) -> int:
    return self._num_channels

  def _get_original_observation(self):
    rpy = self._robot.base_roll_pitch_yaw
    observations = np.zeros(self._num_channels)
    for i, channel in enumerate(self._channels):
      if channel == IMUChannel.ROLL:
        observations[i] = rpy[0]
      elif channel == IMUChannel.PITCH:
        observations[i] = rpy[1]
      elif channel == IMUChannel.YAW:
        observations[i] = rpy[2]
      elif channel == IMUChannel.ROLL_RATE:
        observations[i] = self._robot.base_roll_pitch_yaw_rate[0]
      elif channel == IMUChannel.PITCH_RATE:
        observations[i] = self._robot.base_roll_pitch_yaw_rate[1]
      elif channel == IMUChannel.YAW_RATE:
        observations[i] = self._robot.base_roll_pitch_yaw_rate[2]

    return self._robot.timestamp, np.array(observations, dtype=self._dtype)

  def get_observation(self) -> np.ndarray:
    delayed_observation = super().get_observation()
    if self._noise_generator:
      if callable(self._noise_generator):
        return self._noise_generator(delayed_observation)
      else:
        return self._noise_generator.add_noise(delayed_observation)

    return delayed_observation



import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
os.sys.path.insert(0, parentdir)
import pybullet_data

from pybullet_envs.scene_abstract import Scene
import pybullet


class StadiumScene(Scene):
  zero_at_running_strip_start_line = True  # if False, center of coordinates (0,0,0) will be at the middle of the stadium
  stadium_halflen = 105 * 0.25  # FOOBALL_FIELD_HALFLEN
  stadium_halfwidth = 50 * 0.25  # FOOBALL_FIELD_HALFWID
  stadiumLoaded = 0

  def episode_restart(self, bullet_client):
    self._p = bullet_client
    Scene.episode_restart(self, bullet_client)  # contains cpp_world.clean_everything()
    if (self.stadiumLoaded == 0):
      self.stadiumLoaded = 1

      # stadium_pose = cpp_household.Pose()
      # if self.zero_at_running_strip_start_line:
      #	 stadium_pose.set_xyz(27, 21, 0)  # see RUN_STARTLINE, RUN_RAD constants

      filename = os.path.join(pybullet_data.getDataPath(), "plane_stadium.sdf")
      self.ground_plane_mjcf = self._p.loadSDF(filename)
      #filename = os.path.join(pybullet_data.getDataPath(),"stadium_no_collision.sdf")
      #self.ground_plane_mjcf = self._p.loadSDF(filename)
      #
      for i in self.ground_plane_mjcf:
        self._p.changeDynamics(i, -1, lateralFriction=0.8, restitution=0.5)
        self._p.changeVisualShape(i, -1, rgbaColor=[1, 1, 1, 0.8])
        self._p.configureDebugVisualizer(pybullet.COV_ENABLE_PLANAR_REFLECTION,i)

      #	for j in range(p.getNumJoints(i)):
      #		self._p.changeDynamics(i,j,lateralFriction=0)
      #despite the name (stadium_no_collision), it DID have collision, so don't add duplicate ground


class SinglePlayerStadiumScene(StadiumScene):
  "This scene created by environment, to work in a way as if there was no concept of scene visible to user."
  multiplayer = False


class MultiplayerStadiumScene(StadiumScene):
  multiplayer = True
  players_count = 3

  def actor_introduce(self, robot):
    StadiumScene.actor_introduce(self, robot)
    i = robot.player_n - 1  # 0 1 2 => -1 0 +1
    robot.move_robot(0, i, 0)



import numpy as np
import copy
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf
from pybullet_envs.deep_mimic.learning.normalizer import Normalizer


class TFNormalizer(Normalizer):

  def __init__(self, sess, scope, size, groups_ids=None, eps=0.02, clip=np.inf):
    self.sess = sess
    self.scope = scope
    super().__init__(size, groups_ids, eps, clip)

    with tf.variable_scope(self.scope):
      self._build_resource_tf()
    return

  # initialze count when loading saved values so that things don't change to quickly during updates
  def load(self):
    self.count = self.count_tf.eval()[0]
    self.mean = self.mean_tf.eval()
    self.std = self.std_tf.eval()
    self.mean_sq = self.calc_mean_sq(self.mean, self.std)
    return

  def update(self):
    super().update()
    self._update_resource_tf()
    return

  def set_mean_std(self, mean, std):
    super().set_mean_std(mean, std)
    self._update_resource_tf()
    return

  def normalize_tf(self, x):
    norm_x = (x - self.mean_tf) / self.std_tf
    norm_x = tf.clip_by_value(norm_x, -self.clip, self.clip)
    return norm_x

  def unnormalize_tf(self, norm_x):
    x = norm_x * self.std_tf + self.mean_tf
    return x

  def _build_resource_tf(self):
    self.count_tf = tf.get_variable(dtype=tf.int32,
                                    name='count',
                                    initializer=np.array([self.count], dtype=np.int32),
                                    trainable=False)
    self.mean_tf = tf.get_variable(dtype=tf.float32,
                                   name='mean',
                                   initializer=self.mean.astype(np.float32),
                                   trainable=False)
    self.std_tf = tf.get_variable(dtype=tf.float32,
                                  name='std',
                                  initializer=self.std.astype(np.float32),
                                  trainable=False)

    self.count_ph = tf.get_variable(dtype=tf.int32, name='count_ph', shape=[1])
    self.mean_ph = tf.get_variable(dtype=tf.float32, name='mean_ph', shape=self.mean.shape)
    self.std_ph = tf.get_variable(dtype=tf.float32, name='std_ph', shape=self.std.shape)

    self._update_op = tf.group(self.count_tf.assign(self.count_ph),
                               self.mean_tf.assign(self.mean_ph), self.std_tf.assign(self.std_ph))
    return

  def _update_resource_tf(self):
    feed = {
        self.count_ph: np.array([self.count], dtype=np.int32),
        self.mean_ph: self.mean,
        self.std_ph: self.std
    }
    self.sess.run(self._update_op, feed_dict=feed)
    return



"""This file implements the gym environment of minitaur.

"""

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import math
import time
import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import pybullet
from pybullet_utils import bullet_client as bc
from . import minitaur
import os
import pybullet_data
from . import minitaur_env_randomizer
from pkg_resources import parse_version

NUM_SUBSTEPS = 5
NUM_MOTORS = 8
MOTOR_ANGLE_OBSERVATION_INDEX = 0
MOTOR_VELOCITY_OBSERVATION_INDEX = MOTOR_ANGLE_OBSERVATION_INDEX + NUM_MOTORS
MOTOR_TORQUE_OBSERVATION_INDEX = MOTOR_VELOCITY_OBSERVATION_INDEX + NUM_MOTORS
BASE_ORIENTATION_OBSERVATION_INDEX = MOTOR_TORQUE_OBSERVATION_INDEX + NUM_MOTORS
ACTION_EPS = 0.01
OBSERVATION_EPS = 0.01
RENDER_HEIGHT = 720
RENDER_WIDTH = 960


class MinitaurBulletEnv(gym.Env):
  """The gym environment for the minitaur.

  It simulates the locomotion of a minitaur, a quadruped robot. The state space
  include the angles, velocities and torques for all the motors and the action
  space is the desired motor angle for each motor. The reward function is based
  on how far the minitaur walks in 1000 steps and penalizes the energy
  expenditure.

  """
  metadata = {"render.modes": ["human", "rgb_array"], "video.frames_per_second": 50}

  def __init__(
      self,
      urdf_root=pybullet_data.getDataPath(),
      action_repeat=1,
      distance_weight=1.0,
      energy_weight=0.005,
      shake_weight=0.0,
      drift_weight=0.0,
      distance_limit=float("inf"),
      observation_noise_stdev=0.0,
      self_collision_enabled=True,
      motor_velocity_limit=np.inf,
      pd_control_enabled=False,  #not needed to be true if accurate motor model is enabled (has its own better PD)
      leg_model_enabled=True,
      accurate_motor_model_enabled=True,
      motor_kp=1.0,
      motor_kd=0.02,
      torque_control_enabled=False,
      motor_overheat_protection=True,
      hard_reset=True,
      on_rack=False,
      render=False,
      kd_for_pd_controllers=0.3,
      env_randomizer=minitaur_env_randomizer.MinitaurEnvRandomizer()):
    """Initialize the minitaur gym environment.

    Args:
      urdf_root: The path to the urdf data folder.
      action_repeat: The number of simulation steps before actions are applied.
      distance_weight: The weight of the distance term in the reward.
      energy_weight: The weight of the energy term in the reward.
      shake_weight: The weight of the vertical shakiness term in the reward.
      drift_weight: The weight of the sideways drift term in the reward.
      distance_limit: The maximum distance to terminate the episode.
      observation_noise_stdev: The standard deviation of observation noise.
      self_collision_enabled: Whether to enable self collision in the sim.
      motor_velocity_limit: The velocity limit of each motor.
      pd_control_enabled: Whether to use PD controller for each motor.
      leg_model_enabled: Whether to use a leg motor to reparameterize the action
        space.
      accurate_motor_model_enabled: Whether to use the accurate DC motor model.
      motor_kp: proportional gain for the accurate motor model.
      motor_kd: derivative gain for the accurate motor model.
      torque_control_enabled: Whether to use the torque control, if set to
        False, pose control will be used.
      motor_overheat_protection: Whether to shutdown the motor that has exerted
        large torque (OVERHEAT_SHUTDOWN_TORQUE) for an extended amount of time
        (OVERHEAT_SHUTDOWN_TIME). See ApplyAction() in minitaur.py for more
        details.
      hard_reset: Whether to wipe the simulation and load everything when reset
        is called. If set to false, reset just place the minitaur back to start
        position and set its pose to initial configuration.
      on_rack: Whether to place the minitaur on rack. This is only used to debug
        the walking gait. In this mode, the minitaur's base is hanged midair so
        that its walking gait is clearer to visualize.
      render: Whether to render the simulation.
      kd_for_pd_controllers: kd value for the pd controllers of the motors
      env_randomizer: An EnvRandomizer to randomize the physical properties
        during reset().
    """
    self._time_step = 0.01
    self._action_repeat = action_repeat
    self._num_bullet_solver_iterations = 300
    self._urdf_root = urdf_root
    self._self_collision_enabled = self_collision_enabled
    self._motor_velocity_limit = motor_velocity_limit
    self._observation = []
    self._env_step_counter = 0
    self._is_render = render
    self._last_base_position = [0, 0, 0]
    self._distance_weight = distance_weight
    self._energy_weight = energy_weight
    self._drift_weight = drift_weight
    self._shake_weight = shake_weight
    self._distance_limit = distance_limit
    self._observation_noise_stdev = observation_noise_stdev
    self._action_bound = 1
    self._pd_control_enabled = pd_control_enabled
    self._leg_model_enabled = leg_model_enabled
    self._accurate_motor_model_enabled = accurate_motor_model_enabled
    self._motor_kp = motor_kp
    self._motor_kd = motor_kd
    self._torque_control_enabled = torque_control_enabled
    self._motor_overheat_protection = motor_overheat_protection
    self._on_rack = on_rack
    self._cam_dist = 1.0
    self._cam_yaw = 0
    self._cam_pitch = -30
    self._hard_reset = True
    self._kd_for_pd_controllers = kd_for_pd_controllers
    self._last_frame_time = 0.0
    print("urdf_root=" + self._urdf_root)
    self._env_randomizer = env_randomizer
    # PD control needs smaller time step for stability.
    if pd_control_enabled or accurate_motor_model_enabled:
      self._time_step /= NUM_SUBSTEPS
      self._num_bullet_solver_iterations /= NUM_SUBSTEPS
      self._action_repeat *= NUM_SUBSTEPS

    if self._is_render:
      self._pybullet_client = bc.BulletClient(connection_mode=pybullet.GUI)
    else:
      self._pybullet_client = bc.BulletClient()

    self.seed()
    self.reset()
    observation_high = (self.minitaur.GetObservationUpperBound() + OBSERVATION_EPS)
    observation_low = (self.minitaur.GetObservationLowerBound() - OBSERVATION_EPS)
    action_dim = 8
    action_high = np.array([self._action_bound] * action_dim)
    self.action_space = spaces.Box(-action_high, action_high, dtype=np.float32)
    self.observation_space = spaces.Box(observation_low, observation_high, dtype=np.float32)
    self.viewer = None
    self._hard_reset = hard_reset  # This assignment need to be after reset()

  def set_env_randomizer(self, env_randomizer):
    self._env_randomizer = env_randomizer

  def configure(self, args):
    self._args = args

  def reset(self):
    if self._hard_reset:
      self._pybullet_client.resetSimulation()
      self._pybullet_client.setPhysicsEngineParameter(
          numSolverIterations=int(self._num_bullet_solver_iterations))
      self._pybullet_client.setTimeStep(self._time_step)
      plane = self._pybullet_client.loadURDF("%s/plane.urdf" % self._urdf_root)
      self._pybullet_client.changeVisualShape(plane, -1, rgbaColor=[1, 1, 1, 0.9])
      self._pybullet_client.configureDebugVisualizer(
          self._pybullet_client.COV_ENABLE_PLANAR_REFLECTION, 0)
      self._pybullet_client.setGravity(0, 0, -10)
      acc_motor = self._accurate_motor_model_enabled
      motor_protect = self._motor_overheat_protection
      self.minitaur = (minitaur.Minitaur(pybullet_client=self._pybullet_client,
                                         urdf_root=self._urdf_root,
                                         time_step=self._time_step,
                                         self_collision_enabled=self._self_collision_enabled,
                                         motor_velocity_limit=self._motor_velocity_limit,
                                         pd_control_enabled=self._pd_control_enabled,
                                         accurate_motor_model_enabled=acc_motor,
                                         motor_kp=self._motor_kp,
                                         motor_kd=self._motor_kd,
                                         torque_control_enabled=self._torque_control_enabled,
                                         motor_overheat_protection=motor_protect,
                                         on_rack=self._on_rack,
                                         kd_for_pd_controllers=self._kd_for_pd_controllers))
    else:
      self.minitaur.Reset(reload_urdf=False)

    if self._env_randomizer is not None:
      self._env_randomizer.randomize_env(self)

    self._env_step_counter = 0
    self._last_base_position = [0, 0, 0]
    self._objectives = []
    self._pybullet_client.resetDebugVisualizerCamera(self._cam_dist, self._cam_yaw,
                                                     self._cam_pitch, [0, 0, 0])
    if not self._torque_control_enabled:
      for _ in range(100):
        if self._pd_control_enabled or self._accurate_motor_model_enabled:
          self.minitaur.ApplyAction([math.pi / 2] * 8)
        self._pybullet_client.stepSimulation()
    return self._noisy_observation()

  def seed(self, seed=None):
    self.np_random, seed = seeding.np_random(seed)
    return [seed]

  def _transform_action_to_motor_command(self, action):
    if self._leg_model_enabled:
      for i, action_component in enumerate(action):
        if not (-self._action_bound - ACTION_EPS <= action_component <=
                self._action_bound + ACTION_EPS):
          raise ValueError("{}th action {} out of bounds.".format(i, action_component))
      action = self.minitaur.ConvertFromLegModel(action)
    return action

  def step(self, action):
    """Step forward the simulation, given the action.

    Args:
      action: A list of desired motor angles for eight motors.

    Returns:
      observations: The angles, velocities and torques of all motors.
      reward: The reward for the current state-action pair.
      done: Whether the episode has ended.
      info: A dictionary that stores diagnostic information.

    Raises:
      ValueError: The action dimension is not the same as the number of motors.
      ValueError: The magnitude of actions is out of bounds.
    """
    if self._is_render:
      # Sleep, otherwise the computation takes less time than real time,
      # which will make the visualization like a fast-forward video.
      time_spent = time.time() - self._last_frame_time
      self._last_frame_time = time.time()
      time_to_sleep = self._action_repeat * self._time_step - time_spent
      if time_to_sleep > 0:
        time.sleep(time_to_sleep)
      base_pos = self.minitaur.GetBasePosition()
      camInfo = self._pybullet_client.getDebugVisualizerCamera()
      curTargetPos = camInfo[11]
      distance = camInfo[10]
      yaw = camInfo[8]
      pitch = camInfo[9]
      targetPos = [
          0.95 * curTargetPos[0] + 0.05 * base_pos[0], 0.95 * curTargetPos[1] + 0.05 * base_pos[1],
          curTargetPos[2]
      ]

      self._pybullet_client.resetDebugVisualizerCamera(distance, yaw, pitch, base_pos)
    action = self._transform_action_to_motor_command(action)
    for _ in range(self._action_repeat):
      self.minitaur.ApplyAction(action)
      self._pybullet_client.stepSimulation()

    self._env_step_counter += 1
    reward = self._reward()
    done = self._termination()
    return np.array(self._noisy_observation()), reward, done, {}

  def render(self, mode="rgb_array", close=False):
    if mode != "rgb_array":
      return np.array([])
    base_pos = self.minitaur.GetBasePosition()
    view_matrix = self._pybullet_client.computeViewMatrixFromYawPitchRoll(
        cameraTargetPosition=base_pos,
        distance=self._cam_dist,
        yaw=self._cam_yaw,
        pitch=self._cam_pitch,
        roll=0,
        upAxisIndex=2)
    proj_matrix = self._pybullet_client.computeProjectionMatrixFOV(fov=60,
                                                                   aspect=float(RENDER_WIDTH) /
                                                                   RENDER_HEIGHT,
                                                                   nearVal=0.1,
                                                                   farVal=100.0)
    (_, _, px, _,
     _) = self._pybullet_client.getCameraImage(width=RENDER_WIDTH,
                                               height=RENDER_HEIGHT,
                                               viewMatrix=view_matrix,
                                               projectionMatrix=proj_matrix,
                                               renderer=pybullet.ER_BULLET_HARDWARE_OPENGL)
    rgb_array = np.array(px)
    rgb_array = rgb_array[:, :, :3]
    return rgb_array

  def get_minitaur_motor_angles(self):
    """Get the minitaur's motor angles.

    Returns:
      A numpy array of motor angles.
    """
    return np.array(self._observation[MOTOR_ANGLE_OBSERVATION_INDEX:MOTOR_ANGLE_OBSERVATION_INDEX +
                                      NUM_MOTORS])

  def get_minitaur_motor_velocities(self):
    """Get the minitaur's motor velocities.

    Returns:
      A numpy array of motor velocities.
    """
    return np.array(
        self._observation[MOTOR_VELOCITY_OBSERVATION_INDEX:MOTOR_VELOCITY_OBSERVATION_INDEX +
                          NUM_MOTORS])

  def get_minitaur_motor_torques(self):
    """Get the minitaur's motor torques.

    Returns:
      A numpy array of motor torques.
    """
    return np.array(
        self._observation[MOTOR_TORQUE_OBSERVATION_INDEX:MOTOR_TORQUE_OBSERVATION_INDEX +
                          NUM_MOTORS])

  def get_minitaur_base_orientation(self):
    """Get the minitaur's base orientation, represented by a quaternion.

    Returns:
      A numpy array of minitaur's orientation.
    """
    return np.array(self._observation[BASE_ORIENTATION_OBSERVATION_INDEX:])

  def is_fallen(self):
    """Decide whether the minitaur has fallen.

    If the up directions between the base and the world is larger (the dot
    product is smaller than 0.85) or the base is very low on the ground
    (the height is smaller than 0.13 meter), the minitaur is considered fallen.

    Returns:
      Boolean value that indicates whether the minitaur has fallen.
    """
    orientation = self.minitaur.GetBaseOrientation()
    rot_mat = self._pybullet_client.getMatrixFromQuaternion(orientation)
    local_up = rot_mat[6:]
    pos = self.minitaur.GetBasePosition()
    return (np.dot(np.asarray([0, 0, 1]), np.asarray(local_up)) < 0.85 or pos[2] < 0.13)

  def _termination(self):
    position = self.minitaur.GetBasePosition()
    distance = math.sqrt(position[0]**2 + position[1]**2)
    return self.is_fallen() or distance > self._distance_limit

  def _reward(self):
    current_base_position = self.minitaur.GetBasePosition()
    forward_reward = current_base_position[0] - self._last_base_position[0]
    drift_reward = -abs(current_base_position[1] - self._last_base_position[1])
    shake_reward = -abs(current_base_position[2] - self._last_base_position[2])
    self._last_base_position = current_base_position
    energy_reward = np.abs(
        np.dot(self.minitaur.GetMotorTorques(),
               self.minitaur.GetMotorVelocities())) * self._time_step
    reward = (self._distance_weight * forward_reward - self._energy_weight * energy_reward +
              self._drift_weight * drift_reward + self._shake_weight * shake_reward)
    self._objectives.append([forward_reward, energy_reward, drift_reward, shake_reward])
    return reward

  def get_objectives(self):
    return self._objectives

  def _get_observation(self):
    self._observation = self.minitaur.GetObservation()
    return self._observation

  def _noisy_observation(self):
    self._get_observation()
    observation = np.array(self._observation)
    if self._observation_noise_stdev > 0:
      observation += (
          np.random.normal(scale=self._observation_noise_stdev, size=observation.shape) *
          self.minitaur.GetObservationUpperBound())
    return observation

  if parse_version(gym.__version__) < parse_version('0.9.6'):
    _render = render
    _reset = reset
    _seed = seed
    _step = step



"""Noise generators to simulate noise in sensor / actuator classes."""

import abc
import gin
import numpy as np


class NoiseGenerator(metaclass=abc.ABCMeta):
  """Base class for noise generators."""

  @abc.abstractmethod
  def _get_noise(self, shape, dtype=None):
    """Gets noise as a numpy array in the specified shape and dtype.

    Tensorflow requires the shape and dtype of noise to be correctly specified,
    so the generator needs to know this to produce data of the correct type.

    Args:
      shape: Shape of the returned array.
      dtype: Datatype of returned array (None for default).
    """

  @abc.abstractmethod
  def add_noise(self, data):
    """Adds noise generated by _get_noise to the given data with clipping.

    Args:
      data: Numpy array of data to be modified with noise.
    """


@gin.configurable
class BiasNoise(NoiseGenerator):
  """Adds bias to the data, possibly with clipping."""

  def __init__(self,
               bias=0.0,
               clipping_lower_bound=-np.inf,
               clipping_upper_bound=np.inf):
    """Create a bias noise generator.

    Args:
      bias: Absolute magnitude of bias applied to input.
      clipping_lower_bound: lower bound of add_noise (use -np.inf to ignore).
      clipping_upper_bound: Upper bound of add_noise (use np.inf to ignore).
    """
    self._bias = bias
    self._clipping_lower_bound = clipping_lower_bound
    self._clipping_upper_bound = clipping_upper_bound

  def _get_noise(self, shape, dtype=None):
    """Create bias noise of the given direction and datatype."""
    return np.full(shape, self._bias, dtype)

  def add_noise(self, data):
    """Add bias noise to the given data, clipping to the given range."""
    noise = self._get_noise(data.shape, data.dtype)
    return np.clip(data + noise, self._clipping_lower_bound,
                   self._clipping_upper_bound)


@gin.configurable
class NormalNoise(BiasNoise):
  """Adds Gaussian noise to the data, possibly with clipping."""

  def __init__(self, scale, **kwargs):
    """Create a normal noise generator.

    Args:
      scale: Absolute magnitude of standard deviation of Gaussian noise. Note
        numpy will throw an error if scale < 0.
      **kwargs: Arguments passed to BiasNoise (e.g. bias and clipping).
    """
    super(NormalNoise, self).__init__(**kwargs)
    self._scale = scale

  def _get_noise(self, shape, dtype=None):
    """Create normal noise of the given direction and datatype."""
    return np.random.normal(self._bias, self._scale, shape).astype(dtype)


@gin.configurable
class UniformNoise(NoiseGenerator):
  """Generates uniform noise in the given range."""

  def __init__(self,
               low,
               high,
               clipping_lower_bound=-np.inf,
               clipping_upper_bound=np.inf):
    """Creates a uniform noise generator.

    Args:
      low: the lower bound of the noise.
      high: the higher bound of the noise.
      clipping_lower_bound: lower bound of add_noise (use -np.inf to ignore).
      clipping_upper_bound: Upper bound of add_noise (use np.inf to ignore).
    """
    super().__init__()
    self._low = low
    self._high = high
    self._clipping_lower_bound = clipping_lower_bound
    self._clipping_upper_bound = clipping_upper_bound

  def _get_noise(self, shape, dtype=None):
    """Generates a noise using the given shape and data type."""
    return np.random.uniform(self._low, self._high, shape).astype(dtype)

  def add_noise(self, data):
    """Adds noise to the given data, clipping to the given bound."""
    noise = self._get_noise(data.shape, data.dtype)
    return np.clip(data + noise, self._clipping_lower_bound,
                   self._clipping_upper_bound)


@gin.configurable
class RangeNoise(NormalNoise):
  """Add normally distributed noise in m, applied to hit fractions in (0, 1).

  This enables us to specify range noise in terms of meters of Gaussian noise
  between a maximum and minimum range, but the add_noise is applied as above
  to values expected to be in a hit fraction range of (0, 1) as needed for the
  SimLidarSensor API. Separate methods return noise or noisify data in meters.
  """

  def __init__(self, range_noise_m, max_range_m, min_range_m=0.0, **kwargs):
    """Create a normal noise generator suitable for use in a range scanner.

    Args:
      range_noise_m: Absolute magnitude of standard deviation of Gaussian noise,
        applied to range observation readngs, measured in meters.
      max_range_m: Maximum range in meters of the data, used for clipping.
      min_range_m: Minimum range in meters of the data, used for clipping.
      **kwargs: Other arguments passed to NormalNoise (principally bias).
    """
    # Validate range values.
    if range_noise_m < 0.0:
      raise ValueError("Range noise should not be negative: %r" % range_noise_m)
    if min_range_m >= max_range_m:
      raise ValueError("min_range_m %s must be less than max_range_m %s" %
                       (min_range_m, max_range_m))

    self._range_noise_m = range_noise_m
    self._max_range_m = max_range_m
    self._min_range_m = min_range_m
    self._total_range = max_range_m - min_range_m
    super(RangeNoise, self).__init__(
        scale=range_noise_m / self._total_range,
        clipping_lower_bound=0.0,
        clipping_upper_bound=1.0,
        **kwargs)

  def _get_noise_m(self, shape, dtype=None):
    """Create normal noise of the given direction and datatype, in meters."""
    return self.range_to_m(self._get_noise(shape=shape, dtype=dtype))

  def add_noise_m(self, data):
    """Add normal noise to the given data, scaled in meters."""
    return self.range_to_m(self.add_noise(self.m_to_range(data)))

  def m_to_range(self, data):
    """Scale data in meters to a range of (0, 1)."""
    return (data - self._min_range_m) / self._total_range

  def range_to_m(self, data):
    """Scale data in range of (0, 1) to meters."""
    return data * self._total_range + self._min_range_m


@gin.configurable
class TwistNoise(object):
  """Add normally distributed noise to twist actions.

    Note this is a simplified noise model in action space designed for parity
    with DriveWorld's  r/s/e/drive_models/twist_drive.py;rcl=307540784;l=161.
    This assumes that the TwistNoise will be applied to a twist action which is
    then clipped, as currently done in wheeled_robot_base.py:

    robotics/reinforcement_learning/minitaur/robots/wheeled_robot_base.py;l=533
    # We assume that the velocity clipping would be absorbed in this API.
    if self._action_filter:
      action = self._action_filter.filter(action)

    where action is a linear_velocity, angular_velocity pair, which is clipped
    to limits subsequently by the _compute_kinematic_base_velocity method.
  """

  def __init__(self,
               linear_velocity_noise_stdev_mps: float,
               linear_velocity_noise_max_stdevs: float,
               angular_velocity_noise_stdev_rps: float,
               angular_velocity_noise_max_stdevs: float,
               noise_scaling_cutoff_mps: float = 0.0):
    """Create a normal noise generator suitable for use in a range scanner.

    Supports the API specified in the DriveWorld TwistDrive class:
    robotics/simulation/environments/drive_models/twist_drive.py;l=54

    Args:
      linear_velocity_noise_stdev_mps: One standard deviation of normal noise
        for linear velocity in meters per second.
      linear_velocity_noise_max_stdevs: Max stdevs for linear velocity noise.
        This ensures that the noise values do not spike too crazy.
      angular_velocity_noise_stdev_rps: One standard deviation of normal noise
        for angular velocity in radians per second.
      angular_velocity_noise_max_stdevs: Max stdevs for angular velocity noise.
      noise_scaling_cutoff_mps: If linear velocity is less than this cutoff,
        linear and angular noise are scaled so that zero velocity produces zero
        noise. This enables a robot at rest to remain at rest, while still
        applying reasonable noise values to finite velocities. Angular velocity
        does not contribute to this computation to keep the model simple.
    """
    # Validate range values.
    if linear_velocity_noise_stdev_mps < 0.0:
      raise ValueError("Linear action noise should not be negative: %r" %
                       linear_velocity_noise_stdev_mps)
    if linear_velocity_noise_max_stdevs < 0.0:
      raise ValueError("Maximum linear noise should not be negative: %r" %
                       linear_velocity_noise_max_stdevs)
    if angular_velocity_noise_stdev_rps < 0.0:
      raise ValueError("Angular action noise should not be negative: %r" %
                       angular_velocity_noise_stdev_rps)
    if angular_velocity_noise_max_stdevs < 0.0:
      raise ValueError("Maximum action noise should not be negative: %r" %
                       angular_velocity_noise_max_stdevs)
    if noise_scaling_cutoff_mps < 0.0:
      raise ValueError("Noise scaling cutoff should not be negative: %r" %
                       noise_scaling_cutoff_mps)

    # Save the values to create our noise later.
    self._noise_shape = [
        linear_velocity_noise_stdev_mps, angular_velocity_noise_stdev_rps
    ]
    # The noise clipping is performed using one standard deviation as the unit.
    self._noise_lower_bound = np.array([
        -linear_velocity_noise_max_stdevs * linear_velocity_noise_stdev_mps,
        -angular_velocity_noise_max_stdevs * angular_velocity_noise_stdev_rps
    ])
    self._noise_upper_bound = -self._noise_lower_bound
    self._noise_scaling_cutoff_mps = noise_scaling_cutoff_mps

  def filter(self, action):
    """Filter the linear and angular velocity by adding noise."""
    linear_velocity, angular_velocity = action
    linear_noise, angular_noise = np.clip(
        np.random.normal(0, self._noise_shape, 2), self._noise_lower_bound,
        self._noise_upper_bound)
    if self._noise_scaling_cutoff_mps:
      clipped_velocity = min(abs(linear_velocity),
                             self._noise_scaling_cutoff_mps)
      scaling_factor = clipped_velocity / self._noise_scaling_cutoff_mps
      linear_noise *= scaling_factor
      angular_noise *= scaling_factor

    return (linear_velocity + linear_noise, angular_velocity + angular_noise)



"""A wrapped MinitaurGymEnv with a built-in controller."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import attr
from gym import spaces
import numpy as np
import gin
from pybullet_envs.minitaur.agents.trajectory_generator import tg_inplace
from pybullet_envs.minitaur.envs.utilities import laikago_pose_utils
from pybullet_envs.minitaur.envs.utilities import minitaur_pose_utils

_NUM_LEGS = 4
_LAIKAGO_NUM_ACTIONS = 12
_FREQ_LOWER_BOUND = 0.0
_FREQ_UPPER_BOUND = 3
_DEFAULT_AMPLITUDE_STANCE = -0.02
_DEFAULT_AMPLITUDE_LIFT = 0.9
_DEFAULT_CENTER_EXTENSION = 0
_DEFAULT_STANCE_LIFT_CUTOFF = 2 * np.pi * 0.67
_DEFAULT_RESIDUAL_RANGE = 0.4
_LAIKAGO_KNEE_ACTION_INDEXES = [2, 5, 8, 11]
MINITAUR_INIT_EXTENSION_POS = 2.0
MINITAUR_INIT_SWING_POS = 0.0


@gin.configurable
class PmtgInplaceWrapperEnv(object):
  """A wrapped LocomotionGymEnv with a built-in trajectory generator."""

  def __init__(self,
               gym_env,
               freq_lower_bound=_FREQ_LOWER_BOUND,
               freq_upper_bound=_FREQ_UPPER_BOUND,
               residual_range=_DEFAULT_RESIDUAL_RANGE,
               amplitude_stance=_DEFAULT_AMPLITUDE_STANCE,
               amplitude_lift=_DEFAULT_AMPLITUDE_LIFT,
               center_extension=_DEFAULT_CENTER_EXTENSION,
               stance_lift_cutoff=_DEFAULT_STANCE_LIFT_CUTOFF):
    """Initializes the TG inplace wrapper class.

    Args:
      gym_env: the gym environment to wrap on.
      freq_lower_bound: minimum frequency that the TGs can be propagated at.
      freq_upper_bound: maximum frequency that the TGs can be propagated at.
      residual_range: range of residuals that can be added to tg outputs.
      amplitude_stance: stance amplitude of TG (see tg_inplace.py for details).
      amplitude_lift: swing amplitude of TG (see tg_inplace.py for details).
      center_extension: center extension of TG (see tg_inplace.py for details).
      stance_lift_cutoff: phase cutoff between stance and lift phase (see
        tg_inplace.py for details).
    """
    self._gym_env = gym_env
    self._num_actions = gym_env.robot.num_motors
    self._tg_phases = tg_inplace.reset()
    self._tg_params = dict(
        amplitude_stance=amplitude_stance,
        amplitude_lift=amplitude_lift,
        center_extension=center_extension,
        stance_lift_cutoff=stance_lift_cutoff)

    # Add the action boundaries for delta time, one per integrator.
    action_low = np.hstack(
        ([-residual_range] * self._num_actions, [freq_lower_bound] * _NUM_LEGS))
    action_high = np.hstack(
        ([residual_range] * self._num_actions, [freq_upper_bound] * _NUM_LEGS))
    self.action_space = spaces.Box(action_low, action_high)

    # Set the observation space and boundaries.
    if hasattr(self._gym_env.observation_space, "spaces"):
      self.observation_space = self._gym_env.observation_space
      self.observation_space.spaces["pmtg_inplace"] = spaces.Box(
          -1 * np.ones(2 * _NUM_LEGS), np.ones(2 * _NUM_LEGS))
    else:
      lower_bound = self._gym_env.observation_space.low
      upper_bound = self._gym_env.observation_space.high
      lower_bound = np.hstack((lower_bound, [-1.] * 2 * _NUM_LEGS))
      upper_bound = np.hstack((upper_bound, [1.] * 2 * _NUM_LEGS))
      self.observation_space = spaces.Box(lower_bound, upper_bound)

  def __getattr__(self, attrb):
    return getattr(self._gym_env, attrb)

  def _modify_observation(self, observation):
    if isinstance(observation, dict):
      observation["tg_inplace"] = np.hstack(
          (np.cos(self._tg_phases), np.sin(self._tg_phases)))
      return observation
    else:
      return np.hstack(
          (observation, np.cos(self._tg_phases), np.sin(self._tg_phases)))

  def reset(self, initial_motor_angles=None, reset_duration=1.0):
    """Resets the environment as well as trajectory generators."""
    self._last_real_time = 0
    self._num_step = 0
    self._tg_phases = tg_inplace.reset()
    if self._num_actions == _LAIKAGO_NUM_ACTIONS:
      # Use laikago's init pose as zero action.
      init_pose = np.array(
          attr.astuple(
              laikago_pose_utils.LaikagoPose(
                  abduction_angle_0=laikago_pose_utils
                  .LAIKAGO_DEFAULT_ABDUCTION_ANGLE,
                  hip_angle_0=laikago_pose_utils.LAIKAGO_DEFAULT_HIP_ANGLE,
                  knee_angle_0=laikago_pose_utils.LAIKAGO_DEFAULT_KNEE_ANGLE,
                  abduction_angle_1=laikago_pose_utils
                  .LAIKAGO_DEFAULT_ABDUCTION_ANGLE,
                  hip_angle_1=laikago_pose_utils.LAIKAGO_DEFAULT_HIP_ANGLE,
                  knee_angle_1=laikago_pose_utils.LAIKAGO_DEFAULT_KNEE_ANGLE,
                  abduction_angle_2=laikago_pose_utils
                  .LAIKAGO_DEFAULT_ABDUCTION_ANGLE,
                  hip_angle_2=laikago_pose_utils.LAIKAGO_DEFAULT_HIP_ANGLE,
                  knee_angle_2=laikago_pose_utils.LAIKAGO_DEFAULT_KNEE_ANGLE,
                  abduction_angle_3=laikago_pose_utils
                  .LAIKAGO_DEFAULT_ABDUCTION_ANGLE,
                  hip_angle_3=laikago_pose_utils.LAIKAGO_DEFAULT_HIP_ANGLE,
                  knee_angle_3=laikago_pose_utils.LAIKAGO_DEFAULT_KNEE_ANGLE)))
      self._init_pose = init_pose
      observation = self._gym_env.reset(init_pose, reset_duration)
    else:
      # Use minitaur's init pose as zero action.
      init_pose = np.array(
          attr.astuple(
              minitaur_pose_utils.MinitaurPose(
                  swing_angle_0=MINITAUR_INIT_SWING_POS,
                  swing_angle_1=MINITAUR_INIT_SWING_POS,
                  swing_angle_2=MINITAUR_INIT_SWING_POS,
                  swing_angle_3=MINITAUR_INIT_SWING_POS,
                  extension_angle_0=MINITAUR_INIT_EXTENSION_POS,
                  extension_angle_1=MINITAUR_INIT_EXTENSION_POS,
                  extension_angle_2=MINITAUR_INIT_EXTENSION_POS,
                  extension_angle_3=MINITAUR_INIT_EXTENSION_POS)))
      initial_motor_angles = minitaur_pose_utils.leg_pose_to_motor_angles(
          init_pose)
      observation = self._gym_env.reset(initial_motor_angles, reset_duration)
    return self._modify_observation(observation)

  def step(self, action):
    """Steps the wrapped PMTG inplace environment."""
    time = self._gym_env.get_time_since_reset()

    # Convert the policy's residual actions to motor space.
    if self._num_actions == _LAIKAGO_NUM_ACTIONS:
      action_residual = np.array(
          attr.astuple(
              laikago_pose_utils.LaikagoPose(*(action[0:self._num_actions]))))
    else:
      action_residual = minitaur_pose_utils.leg_pose_to_motor_angles(
          action[0:self._num_actions])

    self._last_real_time = time
    self._tg_phases, tg_extensions = tg_inplace.step(
        self._tg_phases, action[-_NUM_LEGS:], self._gym_env.env_time_step,
        self._tg_params)
    # Convert TG's actions to motor space depending on the robot type.
    if self._num_actions == _LAIKAGO_NUM_ACTIONS:
      # If the legs have 3 DOF, apply extension directly to knee joints
      action_tg_motor_space = np.zeros(self._num_actions)
      for tg_idx, knee_idx in zip(
          range(_NUM_LEGS), _LAIKAGO_KNEE_ACTION_INDEXES):
        action_tg_motor_space[knee_idx] = tg_extensions[tg_idx]
    else:
      # Conversion to motor space for minitaur robot.
      action_tg_motor_space = []
      for idx in range(_NUM_LEGS):
        extend = tg_extensions[idx]
        action_tg_motor_space.extend(
            minitaur_pose_utils.swing_extend_to_motor_angles(idx, 0, extend))
    new_action = action_tg_motor_space + action_residual
    if self._num_actions == _LAIKAGO_NUM_ACTIONS:
      new_action += self._init_pose
    original_observation, reward, done, _ = self._gym_env.step(new_action)

    return self._modify_observation(original_observation), reward, done, _



import pybullet as p
import time
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
planeId = p.loadURDF("plane.urdf", useMaximalCoordinates=False)
cubeId = p.loadURDF("cube_collisionfilter.urdf", [0, 0, 3], useMaximalCoordinates=False)

collisionFilterGroup = 0
collisionFilterMask = 0
p.setCollisionFilterGroupMask(cubeId, -1, collisionFilterGroup, collisionFilterMask)

enableCollision = 1
p.setCollisionFilterPair(planeId, cubeId, -1, -1, enableCollision)

p.setRealTimeSimulation(1)
p.setGravity(0, 0, -10)
while (p.isConnected()):
  time.sleep(1. / 240.)
  p.setGravity(0, 0, -10)



"""A Raibert style controller for Minitaur."""

import collections
import math
import numpy as np

_NUM_MOTORS = 8
_NUM_LEGS = 4

_UPPER_LEG_LEN = 0.112
_LOWER_SHORT_LEG_LEN = 0.199
_LOWER_LONG_LEG_LEN = 0.2315

DIAGONAL_LEG_PAIR_1 = (0, 3)
DIAGONAL_LEG_PAIR_2 = (1, 2)


class BehaviorParameters(
    collections.namedtuple("BehaviorParameters", [
        "stance_duration", "desired_forward_speed", "turning_speed", "standing_height",
        "desired_incline_angle"
    ])):
  __slots__ = ()

  def __new__(cls,
              stance_duration=0.25,
              desired_forward_speed=0.2,
              turning_speed=0,
              standing_height=0.21,
              desired_incline_angle=0):
    return super(BehaviorParameters,
                 cls).__new__(cls, stance_duration, desired_forward_speed, turning_speed,
                              standing_height, desired_incline_angle)


def motor_angles_to_leg_pose(motor_angles):
  leg_pose = np.zeros(_NUM_MOTORS)
  for i in range(_NUM_LEGS):
    leg_pose[i] = 0.5 * (-1)**(i // 2) * (motor_angles[2 * i + 1] - motor_angles[2 * i])
    leg_pose[_NUM_LEGS + i] = 0.5 * (motor_angles[2 * i] + motor_angles[2 * i + 1])
  return leg_pose


def leg_pose_to_motor_angles(leg_pose):
  motor_pose = np.zeros(_NUM_MOTORS)
  for i in range(_NUM_LEGS):
    motor_pose[2 * i] = leg_pose[_NUM_LEGS + i] - (-1)**(i // 2) * leg_pose[i]
    motor_pose[2 * i + 1] = (leg_pose[_NUM_LEGS + i] + (-1)**(i // 2) * leg_pose[i])
  return motor_pose


def leg_pose_to_foot_position(leg_pose):
  """The forward kinematics."""
  l1 = _UPPER_LEG_LEN
  l2 = _LOWER_SHORT_LEG_LEN
  l3 = _LOWER_LONG_LEG_LEN

  ext = leg_pose[1]
  alpha = math.asin(l1 * math.sin(ext) / l2)

  sw = leg_pose[0]
  x = l3 * math.sin(alpha + sw) - l1 * math.sin(ext + sw)
  y = l3 * math.cos(alpha + sw) - l1 * math.cos(ext + sw)

  return (x, -y)


def foot_position_to_leg_pose(foot_position):
  """The inverse kinematics."""
  l1 = _UPPER_LEG_LEN
  l2 = _LOWER_SHORT_LEG_LEN
  l3 = _LOWER_LONG_LEG_LEN

  x = foot_position[0]
  y = foot_position[1]
  assert (y < 0)
  hip_toe_sqr = x**2 + y**2
  cos_beta = (l1 * l1 + l3 * l3 - hip_toe_sqr) / (2 * l1 * l3)
  hip_ankle_sqr = l1 * l1 + l2 * l2 - 2 * l1 * l2 * cos_beta
  hip_ankle = math.sqrt(hip_ankle_sqr)
  cos_ext = -(l1 * l1 + hip_ankle_sqr - l2 * l2) / (2 * l1 * hip_ankle)
  ext = math.acos(cos_ext)

  hip_toe = math.sqrt(hip_toe_sqr)
  cos_theta = (hip_toe_sqr + hip_ankle_sqr - (l3 - l2)**2) / (2 * hip_ankle * hip_toe)

  assert cos_theta > 0
  theta = math.acos(cos_theta)
  sw = math.asin(x / hip_toe) - theta
  return (-sw, ext)


def foot_horizontal_position_to_leg_swing(foot_horizontal_position, leg_extension):
  """Computes the target leg swing.

  Sometimes it is more convenient to plan in the hybrid space.
  """

  l1 = _UPPER_LEG_LEN
  l2 = _LOWER_SHORT_LEG_LEN
  l3 = _LOWER_LONG_LEG_LEN
  ext = leg_extension
  alpha = math.asin(l1 / l2 * math.sin(ext))

  toe_hip_orth = l3 * math.sin(alpha) - l1 * math.sin(ext)
  toe_hip_proj = l3 * math.cos(alpha) - l1 * math.cos(ext)

  theta = math.atan(toe_hip_orth / toe_hip_proj)

  # We may allow theta < 0 for backward foot location.
  # assert theta > 0

  toe_hip_len = math.sqrt(toe_hip_orth**2 + toe_hip_proj**2)

  # Cap the foot horizontal (projected) position so the target leg pose is
  # always feasible.
  foot_position = max(min(toe_hip_len * 0.8, foot_horizontal_position), -toe_hip_len * 0.5)

  sw_and_theta = math.asin(foot_position / toe_hip_len)

  sw = sw_and_theta - theta

  # TODO(tingnan): Fix the sign bug.
  return -sw


def extension_to_ankle_dist(ext):
  l1 = _UPPER_LEG_LEN
  l2 = _LOWER_SHORT_LEG_LEN
  l3 = _LOWER_LONG_LEG_LEN
  alpha = math.asin(l1 / l2 * math.sin(ext))
  return l2 * math.cos(alpha) - l1 * math.cos(ext)


def ankle_dist_to_extension(dist):
  l1 = _UPPER_LEG_LEN
  l2 = _LOWER_SHORT_LEG_LEN
  l3 = _LOWER_LONG_LEG_LEN
  cos_ext = -(l1**2 + dist**2 - l2**2) / (2 * l1 * dist)
  return math.acos(cos_ext)


def generate_swing_trajectory(phase, init_pose, end_pose):
  # Try phase compression
  normalized_phase = math.sqrt(min(phase * 1.5, 1))

  # For swing, we use a linear interpolation:
  sw = (end_pose[0] - init_pose[0]) * normalized_phase + init_pose[0]

  # For extension, we can fit a second order polynomial:
  min_ext = (init_pose[1] + end_pose[1]) / 2 - 0.8
  min_ext = max(min_ext, 0.5)
  phi = 0.7

  min_delta = extension_to_ankle_dist(min_ext)
  init_delta = extension_to_ankle_dist(init_pose[1])
  end_delta = extension_to_ankle_dist(end_pose[1])

  # The polynomial is: a * phi^2 + b * phi + c
  delta_1 = min_delta - init_delta
  delta_2 = end_delta - init_delta
  delta_p = phi * phi - phi

  a = (delta_1 - phi * delta_2) / delta_p

  b = (phi * phi * delta_2 - delta_1) / delta_p

  delta = (a * normalized_phase * normalized_phase + b * normalized_phase + init_delta)

  l1 = _UPPER_LEG_LEN
  l2 = _LOWER_SHORT_LEG_LEN

  delta = min(max(delta, l2 - l1 + 0.01), l2 + l1 - 0.01)

  ext = ankle_dist_to_extension(delta)

  return (sw, ext)


def generate_stance_trajectory(phase, init_pose, end_pose):
  normalized_phase = math.sqrt(phase)
  sw = (end_pose[0] - init_pose[0]) * normalized_phase + init_pose[0]
  ext = end_pose[1]
  return (sw, ext)


class RaibertSwingLegController(object):

  def __init__(self,
               speed_gain=0.05,
               leg_extension_clearance=0.3,
               leg_trajectory_generator=generate_swing_trajectory):
    self._speed_gain = speed_gain
    self._leg_extension_clearance = leg_extension_clearance
    self._leg_trajectory_generator = leg_trajectory_generator

  def get_action(self, raibiert_controller):
    current_speed = raibiert_controller.estimate_base_velocity()
    phase = raibiert_controller.get_phase()

    leg_pose_set = []
    for i in raibiert_controller.swing_set:
      target_foot_horizontal_position = (
          raibiert_controller.behavior_parameters.stance_duration / 2 * current_speed +
          self._speed_gain *
          (current_speed - raibiert_controller.behavior_parameters.desired_forward_speed))

      # Use the swing phase [0, 1] to track the foot. The idea is
      # straightforward:
      # 1) Calculate the target swing and leg extension based on target foot position.
      # 2) Generate a smooth Bezier curve between current leg pose and target pose
      # 3) Find the next leg pose on the curve based on how much time left.

      # 1) Convert the target foot
      target_leg_extension = (raibiert_controller.nominal_leg_extension -
                              self._leg_extension_clearance)
      target_leg_swing = foot_horizontal_position_to_leg_swing(target_foot_horizontal_position,
                                                               leg_extension=target_leg_extension)

      target_leg_pose = (target_leg_swing, target_leg_extension)

      # 2) Generates the curve from the current leg pose to the target leg pose.
      # and Find the next leg pose on the curve based on current swing phase.

      desired_leg_pose = self._leg_trajectory_generator(phase,
                                                        raibiert_controller.swing_start_leg_pose,
                                                        target_leg_pose)

      leg_pose_set.append(desired_leg_pose)

    # 3) adjust the pose with a feedback term to maintain leg height

    return leg_pose_set


class RaibertStanceLegController(object):

  def __init__(self, speed_gain=0.1, leg_trajectory_generator=generate_stance_trajectory):
    self._speed_gain = speed_gain
    self._leg_trajectory_generator = leg_trajectory_generator

  def get_action(self, raibiert_controller):
    phase = raibiert_controller.get_phase()
    current_speed = raibiert_controller.estimate_base_velocity()
    leg_pose_set = []
    for i in raibiert_controller.stance_set:
      desired_forward_speed = (raibiert_controller.behavior_parameters.desired_forward_speed)
      target_foot_position = -(raibiert_controller.behavior_parameters.stance_duration / 2 *
                               current_speed - self._speed_gain *
                               (current_speed - desired_forward_speed))

      target_leg_pose = (foot_horizontal_position_to_leg_swing(
          target_foot_position, leg_extension=raibiert_controller.nominal_leg_extension),
                         raibiert_controller.nominal_leg_extension)

      desired_leg_pose = self._leg_trajectory_generator(phase,
                                                        raibiert_controller.stance_start_leg_pose,
                                                        target_leg_pose)

      leg_pose_set.append(desired_leg_pose)

    return leg_pose_set


class MinitaurRaibertTrottingController(object):
  """A Raibert style controller for trotting gait."""

  def __init__(self,
               robot,
               behavior_parameters=BehaviorParameters(),
               swing_leg_controller=RaibertSwingLegController(),
               stance_leg_controller=RaibertStanceLegController(),
               pose_feedback_controller=None):
    self._time = 0
    self._robot = robot
    self._behavior_parameters = behavior_parameters

    nominal_leg_pose = foot_position_to_leg_pose((0, -self._behavior_parameters.standing_height))
    self._nominal_leg_extension = nominal_leg_pose[1]

    self._swing_leg_controller = swing_leg_controller
    self._stance_leg_controller = stance_leg_controller
    self._pose_feeback_controller = pose_feedback_controller

    # The leg order is FL, RL, FR, RR -> [0, 1, 2, 3]
    self._swing_set = DIAGONAL_LEG_PAIR_1
    self._stance_set = DIAGONAL_LEG_PAIR_2
    self._swing_start_leg_pose = self.get_swing_leg_pose()
    self._stance_start_leg_pose = self.get_stance_leg_pose()

  @property
  def behavior_parameters(self):
    return self._behavior_parameters

  @behavior_parameters.setter
  def behavior_parameters(self, behavior_parameters):
    self._behavior_parameters = behavior_parameters

  @property
  def nominal_leg_extension(self):
    return self._nominal_leg_extension

  @property
  def swing_set(self):
    return self._swing_set

  @property
  def stance_set(self):
    return self._stance_set

  @property
  def swing_start_leg_pose(self):
    return self._swing_start_leg_pose

  @property
  def stance_start_leg_pose(self):
    return self._stance_start_leg_pose

  def _get_average_leg_pose(self, leg_indices):
    """Get the average leg pose."""
    current_leg_pose = motor_angles_to_leg_pose(self._robot.GetMotorAngles())

    # extract the swing leg pose from the current_leg_pose
    leg_pose = []
    for index in leg_indices:
      leg_pose.append([current_leg_pose[index], current_leg_pose[index + _NUM_LEGS]])

    leg_pose = np.array(leg_pose)
    return np.mean(leg_pose, axis=0)

  def get_swing_leg_pose(self):
    """Get the current swing legs' average pose."""
    return self._get_average_leg_pose(self._swing_set)

  def get_stance_leg_pose(self):
    """Get the current stance legs' average pose."""
    return self._get_average_leg_pose(self._stance_set)

  def get_phase(self):
    """Compute the current stance/swing phase."""
    return math.fmod(
        self._time,
        self._behavior_parameters.stance_duration) / self._behavior_parameters.stance_duration

  def update_swing_stance_set(self):
    """Switch the set of swing/stance legs based on timing."""
    swing_stance_phase = math.fmod(self._time, 2 * self._behavior_parameters.stance_duration)
    if swing_stance_phase < self._behavior_parameters.stance_duration:
      return (DIAGONAL_LEG_PAIR_1, DIAGONAL_LEG_PAIR_2)
    return (DIAGONAL_LEG_PAIR_2, DIAGONAL_LEG_PAIR_1)

  def update(self, t):
    self._time = t
    new_swing_set, new_stance_set = self.update_swing_stance_set()

    # If there is a stance/swing switch.
    if new_swing_set[0] is not self._swing_set[0]:
      self._swing_set = new_swing_set
      self._stance_set = new_stance_set

      # Also records the starting pose.
      self._swing_start_leg_pose = self.get_swing_leg_pose()
      self._stance_start_leg_pose = self.get_stance_leg_pose()

  def estimate_base_velocity(self):
    # TODO(tingnan): consider using a sensor fusion.
    stance_leg_pose = self.get_stance_leg_pose()

    delta_sw = stance_leg_pose[0] - self._stance_start_leg_pose[0]

    x, y = leg_pose_to_foot_position(stance_leg_pose)
    toe_hip_len = math.sqrt(x**2 + y**2)
    horizontal_dist = toe_hip_len * delta_sw
    phase = self.get_phase()
    speed = 0 if phase < 0.1 else horizontal_dist / (self._behavior_parameters.stance_duration *
                                                     phase)
    return speed

  def get_swing_leg_action(self):
    return self._swing_leg_controller.get_action(self)

  def get_stance_leg_action(self):
    return self._stance_leg_controller.get_action(self)

  def get_action(self):
    leg_pose = [0] * _NUM_MOTORS
    swing_leg_pose = self.get_swing_leg_action()
    j = 0
    for i in self._swing_set:
      leg_pose[i] = swing_leg_pose[j][0]
      leg_pose[i + _NUM_LEGS] = swing_leg_pose[j][1]
      j += 1

    stance_leg_pose = self.get_stance_leg_action()
    j = 0
    for i in self._stance_set:
      leg_pose[i] = stance_leg_pose[j][0]
      leg_pose[i + _NUM_LEGS] = stance_leg_pose[j][1]
      j += 1

    return leg_pose_to_motor_angles(leg_pose)



"""Abstract base class for environment randomizer."""

import abc


class EnvRandomizerBase(object):
  """Abstract base class for environment randomizer.

  An EnvRandomizer is called in environment.reset(). It will
  randomize physical parameters of the objects in the simulation.
  The physical parameters will be fixed for that episode and be
  randomized again in the next environment.reset().
  """

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def randomize_env(self, env):
    """Randomize the simulated_objects in the environment.

    Args:
      env: The environment to be randomized.
    """
    pass



import pybullet as p
import pybullet_data as pd
import math
import time

p.connect(p.GUI)
p.setAdditionalSearchPath(pd.getDataPath())

textureId = -1

useProgrammatic = 0
useTerrainFromPNG = 1
useDeepLocoCSV = 2
updateHeightfield = False

heightfieldSource = useProgrammatic
import random
random.seed(10)
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING,0)
heightPerturbationRange = 0.05
if heightfieldSource==useProgrammatic:
  numHeightfieldRows = 256
  numHeightfieldColumns = 256
  heightfieldData = [0]*numHeightfieldRows*numHeightfieldColumns 
  for j in range (int(numHeightfieldColumns/2)):
    for i in range (int(numHeightfieldRows/2) ):
      height = random.uniform(0,heightPerturbationRange)
      heightfieldData[2*i+2*j*numHeightfieldRows]=height
      heightfieldData[2*i+1+2*j*numHeightfieldRows]=height
      heightfieldData[2*i+(2*j+1)*numHeightfieldRows]=height
      heightfieldData[2*i+1+(2*j+1)*numHeightfieldRows]=height
      
      
  terrainShape = p.createCollisionShape(shapeType = p.GEOM_HEIGHTFIELD, meshScale=[.05,.05,1], heightfieldTextureScaling=(numHeightfieldRows-1)/2, heightfieldData=heightfieldData, numHeightfieldRows=numHeightfieldRows, numHeightfieldColumns=numHeightfieldColumns)
  terrain  = p.createMultiBody(0, terrainShape)
  p.resetBasePositionAndOrientation(terrain,[0,0,0], [0,0,0,1])

if heightfieldSource==useDeepLocoCSV:
  terrainShape = p.createCollisionShape(shapeType = p.GEOM_HEIGHTFIELD, meshScale=[.5,.5,2.5],fileName = "heightmaps/ground0.txt", heightfieldTextureScaling=128)
  terrain  = p.createMultiBody(0, terrainShape)
  p.resetBasePositionAndOrientation(terrain,[0,0,0], [0,0,0,1])

if heightfieldSource==useTerrainFromPNG:
  terrainShape = p.createCollisionShape(shapeType = p.GEOM_HEIGHTFIELD, meshScale=[.1,.1,24],fileName = "heightmaps/wm_height_out.png")
  textureId = p.loadTexture("heightmaps/gimp_overlay_out.png")
  terrain  = p.createMultiBody(0, terrainShape)
  p.changeVisualShape(terrain, -1, textureUniqueId = textureId)
 
 
p.changeVisualShape(terrain, -1, rgbaColor=[1,1,1,1])


sphereRadius = 0.05
colSphereId = p.createCollisionShape(p.GEOM_SPHERE, radius=sphereRadius)
colBoxId = p.createCollisionShape(p.GEOM_BOX,
                                  halfExtents=[sphereRadius, sphereRadius, sphereRadius])

mass = 1
visualShapeId = -1

link_Masses = [1]
linkCollisionShapeIndices = [colBoxId]
linkVisualShapeIndices = [-1]
linkPositions = [[0, 0, 0.11]]
linkOrientations = [[0, 0, 0, 1]]
linkInertialFramePositions = [[0, 0, 0]]
linkInertialFrameOrientations = [[0, 0, 0, 1]]
indices = [0]
jointTypes = [p.JOINT_REVOLUTE]
axis = [[0, 0, 1]]

for i in range(3):
  for j in range(3):
    for k in range(3):
      basePosition = [
          i * 5 * sphereRadius, j * 5 * sphereRadius, 1 + k * 5 * sphereRadius + 1
      ]
      baseOrientation = [0, 0, 0, 1]
      if (k & 2):
        sphereUid = p.createMultiBody(mass, colSphereId, visualShapeId, basePosition,
                                      baseOrientation)
      else:
        sphereUid = p.createMultiBody(mass,
                                      colBoxId,
                                      visualShapeId,
                                      basePosition,
                                      baseOrientation,
                                      linkMasses=link_Masses,
                                      linkCollisionShapeIndices=linkCollisionShapeIndices,
                                      linkVisualShapeIndices=linkVisualShapeIndices,
                                      linkPositions=linkPositions,
                                      linkOrientations=linkOrientations,
                                      linkInertialFramePositions=linkInertialFramePositions,
                                      linkInertialFrameOrientations=linkInertialFrameOrientations,
                                      linkParentIndices=indices,
                                      linkJointTypes=jointTypes,
                                      linkJointAxis=axis)

      
      p.changeDynamics(sphereUid,
                       -1,
                       spinningFriction=0.001,
                       rollingFriction=0.001,
                       linearDamping=0.0)
      for joint in range(p.getNumJoints(sphereUid)):
        p.setJointMotorControl2(sphereUid, joint, p.VELOCITY_CONTROL, targetVelocity=1, force=10)

p.configureDebugVisualizer(p.COV_ENABLE_RENDERING,1)
p.setGravity(0, 0, -10)
p.setRealTimeSimulation(1)

p.getNumJoints(sphereUid)
for i in range(p.getNumJoints(sphereUid)):
  p.getJointInfo(sphereUid, i)


while (p.isConnected()):
  keys = p.getKeyboardEvents()
  
  if updateHeightfield and heightfieldSource==useProgrammatic:
    for j in range (int(numHeightfieldColumns/2)):
      for i in range (int(numHeightfieldRows/2) ):
        height = random.uniform(0,heightPerturbationRange)#+math.sin(time.time())
        heightfieldData[2*i+2*j*numHeightfieldRows]=height
        heightfieldData[2*i+1+2*j*numHeightfieldRows]=height
        heightfieldData[2*i+(2*j+1)*numHeightfieldRows]=height
        heightfieldData[2*i+1+(2*j+1)*numHeightfieldRows]=height
    #GEOM_CONCAVE_INTERNAL_EDGE may help avoid getting stuck at an internal (shared) edge of the triangle/heightfield.
    #GEOM_CONCAVE_INTERNAL_EDGE is a bit slower to build though.
    #flags = p.GEOM_CONCAVE_INTERNAL_EDGE
    flags = 0
    terrainShape2 = p.createCollisionShape(shapeType = p.GEOM_HEIGHTFIELD, flags = flags, meshScale=[.05,.05,1], heightfieldTextureScaling=(numHeightfieldRows-1)/2, heightfieldData=heightfieldData, numHeightfieldRows=numHeightfieldRows, numHeightfieldColumns=numHeightfieldColumns, replaceHeightfieldIndex = terrainShape)
    

  #print(keys)
  #getCameraImage note: software/TinyRenderer doesn't render/support heightfields!
  #p.getCameraImage(320,200, renderer=p.ER_BULLET_HARDWARE_OPENGL)
  time.sleep(0.01)



"""The common setups for MPC based locoomtion controller environments."""
import time
import gin
import numpy as np

from pybullet_envs.minitaur.agents.baseline_controller import com_velocity_estimator
from pybullet_envs.minitaur.agents.baseline_controller import imu_based_com_velocity_estimator
from pybullet_envs.minitaur.agents.baseline_controller import locomotion_controller
from pybullet_envs.minitaur.agents.baseline_controller import openloop_gait_generator
from pybullet_envs.minitaur.agents.baseline_controller import raibert_swing_leg_controller
from pybullet_envs.minitaur.agents.baseline_controller import torque_stance_leg_controller
#from pybullet_envs.minitaur.envs.env_randomizers import minitaur_push_randomizer
from pybullet_envs.minitaur.envs.env_randomizers import minitaur_push_randomizer
from pybullet_envs.minitaur.robots import robot_config
import pybullet_data as pd

CONFIG_FILE = (pd.getDataPath()+"/configs_v2/base/laikago_reactive.gin")

_MOTOR_KD = [1.0, 2.0, 2.0] * 4
_BODY_HEIGHT = 0.45
_MAX_TIME_SECONDS = 1000000000
_MOTOR_KD = [1.0, 2.0, 2.0] * 4
# TODO(tingnan): This is for tunining the moments of inertia of the model.
# Once we identified the correct value we can remove this.
_SCALE = 4
_INERTIA = (0.07335 * _SCALE, 0, 0, 0, 0.25068 * _SCALE, 0, 0, 0,
            0.25447 * _SCALE)


def load_sim_config(render=True):
  """Builds the environment for the quadruped robot.

  Args:
    render: Enable/disable rendering.
  """
  gin.clear_config(clear_constants=False)
  config_file = CONFIG_FILE
  gin.parse_config_file(config_file)

  # Sim bindings
  # Overwrite a few parameters.

  action_repeat = 4
  gin.bind_parameter("SimulationParameters.num_action_repeat", action_repeat)
  gin.bind_parameter("laikago_v2.Laikago.action_repeat", action_repeat)

  # Control latency is NOT modeled properly for inverse kinematics and
  # jacobians, as we are directly calling the pybullet API. We will try to fix
  # this by loading a separate pybullet instance, set the pose and joint
  # angles which has latency in them, and then run the jacobian/IK.
  gin.bind_parameter("laikago_v2.Laikago.motor_control_mode",
                     robot_config.MotorControlMode.HYBRID)
  # Bump up a bit the adduction/abduction motor d gain for a better tracking.
  gin.bind_parameter("hybrid_motor_model.HybridMotorModel.kd", _MOTOR_KD)
  gin.bind_parameter("SimulationParameters.enable_rendering", render)
  gin.bind_parameter("env_loader.load.wrapper_classes", [])



def add_random_push_config():
  """Adds a random push randomizers to the config."""
  try:
    current_env_randomizers = gin.query_parameter(
        "locomotion_gym_env.LocomotionGymEnv.env_randomizers")

    current_env_randomizers.append(
        minitaur_push_randomizer.MinitaurPushRandomizer(
            horizontal_force_bound=(500, 900),
            vertical_force_bound=(50, 100),
            visualize_perturbation_force=True))
    gin.bind_parameter("locomotion_gym_env.LocomotionGymEnv.env_randomizers",
                       current_env_randomizers)
  except ValueError:
    # No randoimzers bind so far
    gin.bind_parameter("locomotion_gym_env.LocomotionGymEnv.env_randomizers", [
        minitaur_push_randomizer.MinitaurPushRandomizer(
            horizontal_force_bound=(500, 900),
            vertical_force_bound=(50, 100),
            visualize_perturbation_force=True)
    ])


def select_gait(gait_type="fast_trot"):
  """Selects a gait pattern.

  Args:
    gait_type: which gait to use.

  Returns:
    A tuple of (stance_duration, duty_factor, initial_phase)
  """
  # Each gait is composed of stance_duration, duty_factor, and
  # init_phase_full_cycle.
  if gait_type == "fast_trot":
    return [0.25] * 4, [0.6] * 4, [0, 0.5, 0.5, 0]
  elif gait_type == "slow_trot":
    return [0.4] * 4, [0.65] * 4, [0, 0.5, 0.5, 0]
  elif gait_type == "walk":
    return [0.75] * 4, [0.8] * 4, [0.25, 0.75, 0.5, 0]
  else:
    # Means four leg stand for as long as possible.
    return [_MAX_TIME_SECONDS] * 4, [0.99] * 4, [0, 0, 0, 0]


def setup_controller(robot,
                     gait="fast_trot",
                     run_on_robot=False,
                     use_ground_truth_velocity=False):
  """Demonstrates how to create a locomotion controller.

  Args:
    robot: A robot instance.
    gait: The type of gait to use.
    run_on_robot: Whether this controller is running on the real robot or not.
    use_ground_truth_velocity: Whether to use ground truth velocity or velocity
      estimator.

  Returns:
    A locomotion controller.
  """
  desired_speed = (0, 0)
  desired_twisting_speed = 0

  feet_positions = np.array(robot.foot_positions())
  feet_positions[:, 2] = 0

  # Sim and real robots have different mass and contact detection parameters.
  body_weight, contact_force_threshold = (200, 20) if run_on_robot else (215, 0)

  stance_duration, duty_factor, init_phase = select_gait(gait)
  gait_generator = openloop_gait_generator.OpenloopGaitGenerator(
      robot,
      stance_duration=stance_duration,
      duty_factor=duty_factor,
      initial_leg_phase=init_phase,
      contact_detection_force_threshold=contact_force_threshold,
  )
  state_estimator = (
      imu_based_com_velocity_estimator.IMUBasedCOMVelocityEstimator(
          robot,
          contact_detection_threshold=contact_force_threshold,
      ))

  # Use this in sim to test ground truth velocity estimation.
  if use_ground_truth_velocity:
    state_estimator = com_velocity_estimator.COMVelocityEstimator(robot)

  sw_controller = raibert_swing_leg_controller.RaibertSwingLegController(
      robot,
      gait_generator,
      state_estimator,
      desired_speed=desired_speed,
      desired_twisting_speed=desired_twisting_speed,
      desired_height=_BODY_HEIGHT,
      local_hip_positions=feet_positions,
  )
  st_controller = torque_stance_leg_controller.TorqueStanceLegController(
      robot,
      gait_generator,
      state_estimator,
      desired_speed=desired_speed,
      desired_twisting_speed=desired_twisting_speed,
      desired_body_height=_BODY_HEIGHT,
      body_mass=body_weight / 9.8,
      body_inertia=_INERTIA,
  )

  controller = locomotion_controller.LocomotionController(
      robot=robot,
      gait_generator=gait_generator,
      state_estimator=state_estimator,
      swing_leg_controller=sw_controller,
      stance_leg_controller=st_controller,
      clock=robot.GetTimeSinceReset)
  return controller
